from grams.algorithm.data_graph.dg_factory import DGFactory
from grams.algorithm.data_graph.dg_graph import (
    DGNode,
    DGEdge,
    CellNode,
    LiteralValueNode,
    EntityValueNode,
    StatementNode,
    ContextSpan,
    EdgeFlowSource,
    EdgeFlowTarget,
    FlowProvenance,
)
from grams.algorithm.data_graph.dg_config import DGConfigs



from collections import defaultdict
from dataclasses import dataclass
from typing import (
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
from operator import attrgetter
from grams.algorithm.data_graph.dg_graph import DGGraph, DGNode, EntityValueNode
from grams.algorithm.inferences.psl_lib import IDMap

from grams.algorithm.data_graph import CellNode
from grams.algorithm.literal_matchers import TextParser
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEdge,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGNode,
    CGStatementNode,
)
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import (
    WDEntity,
    WDProperty,
    WDQuantityPropertyStats,
    WDEntityLabel,
    WDClass,
)
from kgdata.wikidata.models.wdproperty import WDPropertyDomains, WDPropertyRanges
from sm.misc.fn_cache import CacheMethod


K = TypeVar("K")
V = TypeVar("V")


class StructureFeature:
    """Extract structured features from the candidate graph

    Args:
        idmap: IDMap
        table: LinkedTable
        cg: Candidate graph
        dg: Data graph
        wdentities: Wikidata entities
        wdentity_labels: Wikidata entity labels
        wdclasses: Wikidata classes
        wdprops: Wikidata properties
        wd_num_prop_stats: Wikidata number property stats
        sim_fn: Similarity function
        candidate_types: mapping from a node id (in candidate graph) to its candidate types.
    """

    # list of properties that all classes can be its domain.
    # these properties are typically not included in the databases to
    # keep the database size reasonable
    NO_DOMAIN_PROPS = {"P31", "P279", "P1647"}
    NO_RANGE_PROPS = {"P31", "P279", "P1647"}

    def __init__(
        self,
        idmap: IDMap,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        wdentities: Mapping[str, WDEntity],
        wdentity_labels: Mapping[str, WDEntityLabel],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
        wdprop_domains: Optional[Mapping[str, WDPropertyDomains]],
        wdprop_ranges: Optional[Mapping[str, WDPropertyRanges]],
        wd_num_prop_stats: Mapping[str, WDQuantityPropertyStats],
        sim_fn: Optional[Callable[[str, str], float]],
        candidate_types: Dict[str, List[str]],
    ):
        self.idmap = idmap
        self.table = table
        self.cg = cg
        self.dg = dg
        self.wdentities = wdentities
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.wdprop_domains = wdprop_domains
        self.wdprop_ranges = wdprop_ranges
        self.wd_num_prop_stats = wd_num_prop_stats

        self.cg_nodes = self.cg.nodes()
        self.cg_edges = self.cg.edges()
        self.candidate_types = candidate_types

    @staticmethod
    def get_relations(cg: CGGraph) -> List[Tuple[CGStatementNode, CGEdge, CGEdge]]:
        rels = []
        for s in cg.iter_nodes():
            if not isinstance(s, CGStatementNode):
                continue
            (inedge,) = cg.in_edges(s.id)
            for outedge in cg.out_edges(s.id):
                rels.append((s, inedge, outedge))
        return rels

    def extract_features(self, features: List[str]) -> Dict[str, list]:
        rels = self.get_relations(self.cg)

        need_rel_feats = {"REL_PROP", "REL_QUAL", "REL", "STATEMENT_PROPERTY"}
        feat_data = {}
        for feat in features:
            fn = getattr(self, feat)
            if feat in need_rel_feats:
                feat_data[feat] = fn(rels)
            else:
                feat_data[feat] = fn()
        return feat_data

    def REL(
        self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]
    ) -> List[Tuple[str, str, str, str]]:
        """Extract relationships in the candidate graph"""
        idmap = self.idmap
        output = []
        for s, inedge, outedge in rels:
            output.append(
                (
                    idmap.m(inedge.source),
                    idmap.m(outedge.target),
                    idmap.m(s.id),
                    idmap.m(outedge.predicate),
                )
            )
        return output

    def STATEMENT_PROPERTY(
        self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]
    ) -> List[Tuple[str, str]]:
        """Extract relationships in the candidate graph"""
        idmap = self.idmap
        output = []
        for s, inedge, outedge in rels:
            if inedge.predicate == outedge.predicate:
                output.append((idmap.m(s.id), idmap.m(outedge.predicate)))
        return output

    def REL_PROP(
        self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]
    ) -> List[Tuple[str, str, str, str]]:
        idmap = self.idmap
        output = []
        for s, inedge, outedge in rels:
            if inedge.predicate == outedge.predicate:
                output.append(
                    (
                        idmap.m(inedge.source),
                        idmap.m(outedge.target),
                        idmap.m(s.id),
                        idmap.m(outedge.predicate),
                    )
                )
        return output

    def REL_QUAL(
        self, rels: List[Tuple[CGStatementNode, CGEdge, CGEdge]]
    ) -> List[Tuple[str, str, str, str]]:
        idmap = self.idmap
        output = []
        for s, inedge, outedge in rels:
            if inedge.predicate != outedge.predicate:
                output.append(
                    (
                        idmap.m(inedge.source),
                        idmap.m(outedge.target),
                        idmap.m(s.id),
                        idmap.m(outedge.predicate),
                    )
                )
        return output

    def TYPE(self) -> List[Tuple[str, str]]:
        """Extract types of nodes in the candidate graph"""
        idmap = self.idmap
        return [
            (idmap.m(uid), idmap.m(type))
            for uid, types in self.candidate_types.items()
            for type in types
        ]

    def HAS_TYPE(self) -> List[Union[Tuple[str], Tuple[str, float]]]:
        idmap = self.idmap
        lst: List[Union[Tuple[str], Tuple[str, float]]] = [
            (idmap.m(uid),) for uid in self.candidate_types.keys()
        ]
        for node in self.cg_nodes:
            if isinstance(node, CGEntityValueNode):
                lst.append((idmap.m(node.id), 1.0))
            elif isinstance(node, CGLiteralValueNode):
                lst.append((idmap.m(node.id), 0.0))
            elif isinstance(node, CGColumnNode) and node.id not in self.candidate_types:
                lst.append((idmap.m(node.id), 0.0))
        return lst

    def COLUMN(self) -> List[Tuple[str]]:
        idmap = self.idmap
        return [(idmap.m(u.id),) for u in self.cg_nodes if isinstance(u, CGColumnNode)]

    def STATEMENT(self) -> List[Tuple[str]]:
        """Extract nodes that are statements in CG"""
        idmap = self.idmap
        return [
            (idmap.m(u.id),) for u in self.cg_nodes if isinstance(u, CGStatementNode)
        ]

    def NOT_STATEMENT(self) -> List[Tuple[str]]:
        """Extract nodes that are not statements in CG"""
        idmap = self.idmap
        return [
            (idmap.m(u.id),)
            for u in self.cg_nodes
            if not isinstance(u, CGStatementNode)
        ]

    def SUB_PROP(self) -> List[Tuple[str, str]]:
        """Extract subproperty of relationship for properties used in the candidate graph"""
        idmap = self.idmap
        props = self.get_props()
        return [
            (idmap.m(prop.id), idmap.m(sub.id))
            for prop in props.values()
            for sub in props.values()
            if sub.id in prop.ancestors
        ]

    def SUB_TYPE(self) -> List[Tuple[str, str]]:
        """Extract subclass of relationship of types used in the candidate graph"""
        idmap = self.idmap
        classes = self.get_classes()
        return [
            (idmap.m(cls.id), idmap.m(parent.id))
            for cls in classes.values()
            for parent in classes.values()
            if parent.id in cls.ancestors
        ]

    def OBJECT_PROPERTY(self) -> List[Tuple[str]]:
        "Extract object properties used in CG"
        props = self.get_props()
        idmap = self.idmap
        return [(idmap.m(p.id),) for p in props.values() if p.is_object_property()]

    def DATA_PROPERTY(self) -> List[Tuple[str]]:
        "Extract data properties used in CG"
        props = self.get_props()
        idmap = self.idmap
        return [(idmap.m(p.id),) for p in props.values() if p.is_data_property()]

    def PROPERTY_DOMAIN(self) -> List[Tuple[str, str]]:
        prop_domains = self.get_prop_domains()
        u2props = {}
        out = set()
        idmap = self.idmap

        for e in self.cg_edges:
            u = self.cg.get_node(e.source)
            if isinstance(u, CGStatementNode):
                # u is the statement node, so we don't need to care as we are only interested in domains
                continue

            if isinstance(u, CGEntityValueNode):
                # always satisfied
                continue

            if u.id not in u2props:
                u2props[u.id] = []
            u2props[u.id].append(e.predicate)

        for uid, props in u2props.items():
            for prop in props:
                if prop in self.NO_DOMAIN_PROPS:
                    for type in self.candidate_types.get(uid, []):
                        out.add((idmap.m(prop), idmap.m(type)))
                else:
                    domains = prop_domains[prop]
                    for type in self.candidate_types.get(uid, []):
                        if type in domains:
                            out.add((idmap.m(prop), idmap.m(type)))
        return list(out)

    def PROPERTY_RANGE(self) -> List[Tuple[str, str]]:
        prop_ranges = self.get_prop_ranges()
        props = self.get_props()

        v2props = {}
        out = set()
        idmap = self.idmap

        for e in self.cg_edges:
            v = self.cg.get_node(e.target)
            if (
                isinstance(v, CGStatementNode)
                or isinstance(v, CGEntityValueNode)
                or not props[e.predicate].is_object_property()
            ):
                # v is the statement node, so we don't need to care as we are only interested in ranges
                # v is an entity, it is always satisfied
                # e is not an object property, we don't have range
                continue

            if v.id not in v2props:
                v2props[v.id] = []
            v2props[v.id].append(e.predicate)

        for vid, props in v2props.items():
            for prop in props:
                if prop in self.NO_RANGE_PROPS:
                    for type in self.candidate_types.get(vid, []):
                        out.add((idmap.m(prop), idmap.m(type)))
                else:
                    ranges = prop_ranges[prop]
                    for type in self.candidate_types.get(vid, []):
                        if type in ranges:
                            out.add((idmap.m(prop), idmap.m(type)))
        return list(out)

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def get_props(self) -> Dict[str, WDProperty]:
        """Get properties used in the candidate graph"""
        prop_ids = {edge.predicate for edge in self.cg_edges}
        return {prop_id: self.wdprops[prop_id] for prop_id in prop_ids}

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def get_classes(self) -> Dict[str, WDClass]:
        """Get classes used in the candidate graph"""
        class_ids = {type for types in self.candidate_types.values() for type in types}
        return {class_id: self.wdclasses[class_id] for class_id in class_ids}

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def get_prop_domains(self) -> Dict[str, WDPropertyDomains]:
        prop_ids = {edge.predicate for edge in self.cg_edges}
        assert self.wdprop_domains is not None, "Property domains not provided"
        return {
            prop_id: self.wdprop_domains[prop_id]
            for prop_id in prop_ids
            if prop_id not in self.NO_DOMAIN_PROPS
        }

    @CacheMethod.cache(CacheMethod.as_is_posargs)
    def get_prop_ranges(self) -> Dict[str, WDPropertyRanges]:
        props = self.get_props()
        assert self.wdprop_ranges is not None, "Property ranges not provided"
        return {
            prop_id: self.wdprop_ranges[prop_id]
            for prop_id, prop in props.items()
            if prop_id not in self.NO_RANGE_PROPS and prop.is_object_property()
        }



from dataclasses import dataclass
from typing import Mapping

from kgdata.wikidata.models import WDEntity


@dataclass
class LiteralMatchKit:
    """Storing objects that may be needed for matching"""

    wdentities: Mapping[str, WDEntity]



from collections import defaultdict
from operator import itemgetter
from typing import Dict, List, Tuple
from grams.algorithm.candidate_graph.cg_graph import CGColumnNode, CGGraph
from grams.algorithm.data_graph.dg_graph import DGGraph
from grams.inputs.linked_table import LinkedTable
from graph.retworkx.api import digraph_all_simple_paths

import sm.misc as M
import networkx as nx


class KnownTargets:
    """Post-processing that selects relationships and types of given columns"""

    instance = None

    def __init__(
        self,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        edge_probs: Dict[Tuple[str, str, str], float],
        cta_probs: Dict[int, Dict[str, float]],
        threshold: float,
        cpa_targets: Dict[str, List[Tuple[int, int]]],
        cta_targets: Dict[str, List[int]],
    ):
        self.table = table
        self.cg = cg
        self.dg = dg
        self.edge_probs = edge_probs
        self.cta_probs = cta_probs
        self.threshold = threshold

        self.cpa_targets = cpa_targets
        self.cta_targets = cta_targets

    def get_result(self):
        column2node = {
            u.column: u.id for u in self.cg.iter_nodes() if isinstance(u, CGColumnNode)
        }

        selected_edges = {}

        edge_probs = {e: p for e, p in self.edge_probs.items() if p >= self.threshold}
        for sci, tci in self.cpa_targets.get(self.table.id, []):
            uid = column2node[sci]
            vid = column2node[tci]

            if not self.cg.has_node(uid) or not self.cg.has_node(vid):
                continue

            paths = digraph_all_simple_paths(self.cg, uid, vid, cutoff=2)
            paths = [
                p
                for p in paths
                if (p[0].source, p[0].target, p[0].predicate) in edge_probs
                and (p[1].source, p[1].target, p[1].predicate) in edge_probs
            ]
            if len(paths) == 0:
                continue

            e1, e2 = max(
                paths,
                key=lambda p: edge_probs[p[0].source, p[0].target, p[0].predicate]
                + edge_probs[p[1].source, p[1].target, p[1].predicate],
            )

            for e in [e1, e2]:
                triple = e.source, e.target, e.predicate
                selected_edges[triple] = self.edge_probs[triple]

        pred_cpa = self.cg.subgraph_from_edge_triples(selected_edges.keys())
        pred_cta = {
            ci: max(self.cta_probs[ci].items(), key=itemgetter(1))[0]
            for ci in self.cta_targets.get(self.table.id, [])
            if ci in self.cta_probs
        }

        return pred_cpa, pred_cta



import os
from pathlib import Path

from omegaconf import OmegaConf

PKG_DIR = Path(os.path.abspath(__file__)).parent
ROOT_DIR = PKG_DIR.parent
if "DATA_DIR" not in os.environ:
    DATA_DIR = ROOT_DIR / "data"
else:
    DATA_DIR = Path(os.environ["DATA_DIR"])
DEFAULT_CONFIG = OmegaConf.load(PKG_DIR / "grams.yaml")



from typing import Tuple, cast
from grams.algorithm.literal_matchers.text_parser import ParsedTextRepr
from kgdata.wikidata.models import WDValueMonolingualText, WDValueString
import rltk
from grams.algorithm.literal_matchers.types import LiteralMatchKit


def monolingual_exact_test(
    kgval: WDValueMonolingualText, val: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgval: the value in KG
        val: the value in the cell
        kit: objects that may be needed for matching

    Returns:
        a tuple of (whether it's matched, confidence)
    """
    kgtext = kgval.value
    if val.normed_string == kgtext["text"].strip():
        return True, 1.0
    return False, 0.0


def string_exact_test(
    kgval: WDValueString, val: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgval: the value in KG
        val: the value in the cell
        kit: objects that may be needed for matching

    Returns:
        a tuple of (whether it's matched, confidence)
    """
    kgtext = kgval.value
    if val.normed_string == kgtext.strip():
        return True, 1.0
    return False, 0.0


def monolingual_similarity_test(
    kgval: WDValueMonolingualText, val: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgval: the value in KG
        val: the value in the cell
        kit: objects that may be needed for matching

    Returns:
        a tuple of (whether it's matched, confidence)
    """
    kgtext = kgval.value
    return string_match_similarity(kgtext["text"].strip(), val.normed_string)


def string_similarity_test(
    kgval: WDValueString, val: ParsedTextRepr, kit: LiteralMatchKit
) -> Tuple[bool, float]:
    """Compare if the value in KG matches with value in the cell

    Args:
        kgval: the value in KG
        val: the value in the cell
        kit: objects that may be needed for matching

    Returns:
        a tuple of (whether it's matched, confidence)
    """
    kgtext = kgval.value
    return string_match_similarity(kgtext.strip(), val.normed_string)


def string_match_similarity(s1: str, s2: str) -> Tuple[bool, float]:
    """Compare if two strings are similar

    Args:
        s1: the value in KG
        s2: the value in the cell

    Returns:
        a tuple of (whether it's matched, confidence)
    """
    if s1 == s2:
        return True, 1.0
    # calculate edit distance
    distance = rltk.levenshtein_distance(s1, s2)
    if distance <= 1:
        return True, 0.95
    elif distance > 1 and distance / max(1, min(len(s1), len(s2))) < 0.03:
        return True, 0.85
    return False, 0.0



import glob
import re
from dataclasses import dataclass
from pathlib import Path
from typing import List

import click
from omegaconf import OmegaConf

import grams.inputs as I
import sm.misc as M
from grams.config import PKG_DIR, ROOT_DIR
from grams.main import GRAMS
from tqdm.auto import tqdm


@dataclass
class IOFile:
    infile: Path
    outfile: Path


@click.command()
@click.option(
    "-i",
    "--infiles",
    required=True,
    help="""path to input tables. You can use a combination of wildcard (`*`) or named wildcard wrapped by curly brackets (e.g., `{name}` or `{group1}`), which behaves as wildcard. The named wildcard can be used in `outfiles` to name the output file. The list of input tables is retrieved by glob.glob (unix style pathname pattern expansion). Examples:
    - `/tables/{filename}.json`
    - `/tables/{filename}.csv`
""",
)
@click.option(
    "-o",
    "--outfiles",
    help="""path to the output annotations. You can only use named wildcard defined in `infiles`. Examples:
    - `/outputs/{filename}/version.01.json`""",
)
@click.option(
    "-d",
    "--data_dir",
    default=ROOT_DIR / "data",
    help="pass through option to GRAMS.data_dir",
)
@click.option(
    "-p",
    "--proxy",
    is_flag=True,
    default=False,
    help="pass through option to GRAMS.proxy",
)
@click.option(
    "--cfg_file",
    default=str(PKG_DIR / "grams.yaml"),
    help="cfg_file contains configuration of GRAMS",
)
@click.option(
    "-r", "--viz", is_flag=True, default=False, help="visualize the annotated models"
)
@click.option(
    "-v", "--verbose", is_flag=True, default=False, help="print the annotating progress"
)
def cli(
    infiles: str,
    outfiles: str,
    data_dir: str,
    proxy: bool,
    cfg_file: str,
    viz: bool,
    verbose: bool,
):
    """Annotate tables using GRAMS

    \b
    Args:
        infiles: path to input tables. You can use a combination of wildcard (`*`) or named wildcard wrapped by curly brackets (e.g., `{name}` or `{group1}`), which behaves as wildcard. The named wildcard can be used in `outfiles` to name the output file. The list of input tables is retrieved by glob.glob (unix style pathname pattern expansion). Examples:
            - `/tables/{filename}.json`
            - `/tables/{filename}.csv`
        outfiles: path to the output annotations. You can only use named wildcard defined in `infiles`. Examples:
            - `/outputs/{filename}/version.01.json`
        data_dir: pass through option to GRAMS.data_dir
        proxy: pass through option to GRAMS.proxy
        cfg_file: cfg_file contains configuration of GRAMS
        viz: visualize the annotated models
        verbose: print the annotating progress
    """
    cfg = OmegaConf.load(cfg_file)
    io_files = io_parser(infiles=infiles, outfiles=outfiles)
    tables = []

    for io_file in io_files:
        if io_file.infile.name.endswith(".csv"):
            tbl = I.LinkedTable.from_csv_file(io_file.infile)
        elif io_file.infile.name.endswith(".json"):
            tbl = I.LinkedTable.from_dict(M.deserialize_json(io_file.infile))
        else:
            raise NotImplementedError("%s" % io_file.infile.suffix)
        io_file.outfile.parent.mkdir(exist_ok=True, parents=True)
        tables.append(tbl)

    grams = GRAMS(data_dir=data_dir, cfg=cfg, proxy=proxy)

    for io_file, tbl in tqdm(
        zip(io_files, tables), total=len(tables), disable=not verbose
    ):
        annotation = grams.annotate(tbl)
        M.serialize_json(
            {"semantic_models": [annotation.sm.to_dict()]}, io_file.outfile
        )
        if viz:
            annotation.sm.draw(io_file.outfile.parent / (io_file.outfile.stem + ".png"))


def io_parser(infiles: str, outfiles: str) -> List[IOFile]:
    # convert infile pattern torn
    infile_glob_ptn = re.sub("{([a-zA-Z0-9_]+)}", "*", infiles)
    # get a regex pattern that extract groups from input path to format the outfile pattern
    infile_re_ptn = re.sub("{([a-zA-Z0-9_]+)}", r"(?P<\g<1>>[a-zA-Z0-9_]+)", infiles)

    outputs = []
    for infile in sorted(glob.glob(infile_glob_ptn)):
        m = re.match(infile_re_ptn, infile)
        assert m is not None
        outfile = outfiles.format(**m.groupdict())
        outputs.append(IOFile(infile=Path(infile), outfile=Path(outfile)))
    return outputs


if __name__ == "__main__":
    cli()



from grams.algorithm.literal_matchers import TextParser, LiteralMatch
from grams.inputs.linked_table import LinkedTable
import networkx as nx
from typing import Dict, Mapping, Optional, Set, Union, cast
from kgdata.wikidata.models import WDEntity, WDProperty, WDClass
from grams.algorithm.data_graph.dg_graph import (
    DGGraph,
    CellNode,
    ContextSpan,
    DGEdge,
    DGNode,
    DGPath,
    DGPathNode,
    DGPathEdge,
    EdgeFlowSource,
    EdgeFlowTarget,
    EntityValueNode,
    DGPathNodeStatement,
    DGPathNodeEntity,
    DGPathNodeLiteralValue,
    DGPathExistingNode,
    LiteralValueNode,
    Span,
    StatementNode,
)
from grams.algorithm.data_graph.dg_inference import KGInference
from grams.algorithm.data_graph.dg_pruning import DGPruning
from grams.algorithm.data_graph.dg_config import DGConfigs
from grams.algorithm.kg_index import KGObjectIndex
from tqdm import tqdm
import sm.misc as M


class DGFactory:
    def __init__(
        self, wdentities: Mapping[str, WDEntity], wdprops: Mapping[str, WDProperty]
    ):
        self.wdentities = wdentities
        self.wdprops = wdprops
        self.textparser = TextParser()
        self.literal_match = LiteralMatch(wdentities)

        if DGConfigs.USE_KG_INDEX:
            self._path_object_search = self._path_object_search_v2
        else:
            self._path_object_search = self._path_object_search_v1

    def create_dg(
        self,
        table: LinkedTable,
        kg_object_index: KGObjectIndex,
        ignore_columns: Optional[Set[int]] = None,
        max_n_hop: int = 2,
        verbose: bool = False,
    ):
        """Build data graph from a linked table"""
        dg = DGGraph()
        context_node_id = None

        for ci, col in enumerate(table.table.columns):
            for ri, val in enumerate(col.values):
                # =====================================================
                # NOTE: old code
                # cell_qnodes = set()
                # cell_qnode_spans = {}
                # for link in table.links[ri][ci]:
                #     if link.entity_id is not None:
                #         cell_qnodes.add(link.entity_id)
                #         if link.entity_id not in cell_qnode_spans:
                #             cell_qnode_spans[link.entity_id] = []
                #         cell_qnode_spans[link.entity_id].append(Span(link.start, link.end))
                # assert all(len(spans) == len(set(spans)) for spans in cell_qnode_spans.values())
                # TODO: new code, doesn't handle the qnode_spans correctly
                cell_qnodes = {
                    candidate.entity_id
                    for link in table.links[ri][ci]
                    for candidate in link.candidates
                }
                cell_qnode_spans = {}
                for link in table.links[ri][ci]:
                    # TODO: old code
                    # if link.entity_id is not None or len(link.candidates) > 0:
                    #     if len(link.candidates) > 0:
                    #         tmpid = link.candidates[0].entity_id
                    #     else:
                    #         tmpid = link.entity_id
                    #     if tmpid not in cell_qnode_spans:
                    #         cell_qnode_spans[tmpid] = []
                    #     cell_qnode_spans[tmpid].append(Span(link.start, link.end))
                    if len(link.candidates) > 0:
                        tmpid = link.candidates[0].entity_id
                        if tmpid not in cell_qnode_spans:
                            cell_qnode_spans[tmpid] = []
                        cell_qnode_spans[tmpid].append(Span(link.start, link.end))
                assert all(
                    len(spans) == len(set(spans)) for spans in cell_qnode_spans.values()
                )
                # =====================================================

                node = CellNode(
                    id=f"{ri}-{ci}",
                    value=val,
                    column=ci,
                    row=ri,
                    entity_ids=list(cell_qnodes),
                    entity_spans=cell_qnode_spans,
                )
                dg.add_node(node)

        if DGConfigs.USE_CONTEXT and table.context.page_entity_id is not None:
            assert table.context.page_title is not None
            context_node_id = DGPathNodeEntity(table.context.page_entity_id).get_id()
            node = EntityValueNode(
                id=context_node_id,
                qnode_id=table.context.page_entity_id,
                context_span=ContextSpan(
                    text=table.context.page_title,
                    span=Span(0, len(table.context.page_title)),
                ),
            )
            dg.add_node(node)

        # find all paths
        n_rows = len(table.table.columns[0].values)
        kg_path_discovering_tasks = []
        new_paths = []
        for ri in range(n_rows):
            for ci, col in enumerate(table.table.columns):
                if ignore_columns is not None and ci in ignore_columns:
                    continue

                u = dg.get_node(f"{ri}-{ci}")
                for cj in range(ci + 1, len(table.table.columns)):
                    if ignore_columns is not None and cj in ignore_columns:
                        continue

                    v = dg.get_node(f"{ri}-{cj}")
                    kg_path_discovering_tasks.append((u, v))
                if context_node_id is not None and dg.has_node(context_node_id):
                    kg_path_discovering_tasks.append((u, dg.get_node(context_node_id)))

        for u, v in (
            tqdm(kg_path_discovering_tasks, desc="KG searching")
            if verbose
            else kg_path_discovering_tasks
        ):
            new_paths += self.kg_path_discovering(
                kg_object_index, dg, u, v, max_n_hop=max_n_hop
            )

        # add paths to the graph
        for path in new_paths:
            curr_nodeid = path.sequence[0].id
            tmp_path = [curr_nodeid]
            for i in range(1, len(path.sequence), 2):
                prop: DGPathEdge = path.sequence[i]
                value: DGPathNode = path.sequence[i + 1]

                if isinstance(value, DGPathExistingNode):
                    nodeid = value.id
                else:
                    nodeid = value.get_id()
                    if not dg.has_node(nodeid):
                        if isinstance(value, DGPathNodeStatement):
                            dg.add_node(
                                StatementNode(
                                    id=nodeid,
                                    qnode_id=value.qnode_id,
                                    predicate=value.predicate,
                                    is_in_kg=True,
                                )
                            )
                        elif isinstance(value, DGPathNodeEntity):
                            dg.add_node(
                                EntityValueNode(
                                    id=nodeid,
                                    qnode_id=value.qnode_id,
                                    context_span=None,
                                ),
                            )
                        else:
                            dg.add_node(
                                LiteralValueNode(
                                    id=nodeid, value=value.value, context_span=None
                                ),
                            )

                if dg.has_edge_between_nodes(curr_nodeid, nodeid, prop.value):
                    edge = dg.get_edge_between_nodes(curr_nodeid, nodeid, prop.value)
                else:
                    edge = DGEdge(
                        source=curr_nodeid,
                        target=nodeid,
                        predicate=prop.value,
                        is_qualifier=prop.is_qualifier,
                    )
                    dg.add_edge(edge)

                tmp_path.append(edge)
                tmp_path.append(nodeid)
                curr_nodeid = nodeid

            # path format: u - e - (s) - e - v - e - (s2) - e - v2
            for i in range(2, len(tmp_path), 4):
                u_node_id = tmp_path[i - 2]
                u_edge = tmp_path[i - 1]
                snode = dg.get_node(tmp_path[i])
                assert isinstance(snode, StatementNode)
                v_edge = tmp_path[i + 1]
                v_node_id = tmp_path[i + 2]

                edge_source = EdgeFlowSource(u_node_id, u_edge.predicate)
                edge_target = EdgeFlowTarget(v_node_id, v_edge.predicate)
                snode.track_provenance(
                    edge_source, edge_target, [path.sequence[i].provenance]
                )

        # DEBUG code
        # for uid, udata in dg.nodes(data=True):
        #     u = udata['data']
        #     if not u.is_cell:
        #         continue
        #     if u.column == 0:
        #         for _, sid, eid, edata in dg.out_edges(uid, data=True, keys=True):
        #             for _1, vid, e2, e2data in dg.out_edges(sid, data=True, keys=True):
        #                 v = dg.nodes[vid]['data']
        #                 if v.is_cell and v.column == 1:
        #                     print(uid, sid, vid, eid, e2)

        KGInference(
            dg, self.wdentities, self.wdprops
        ).infer_subproperty().kg_transitive_inference()

        # pruning unnecessary paths
        if DGConfigs.PRUNE_REDUNDANT_ENT:
            DGPruning(dg).prune_hidden_entities()

        M.log("grams", data_graph=dg)
        return dg

    def kg_path_discovering(
        self,
        kg_object_index: KGObjectIndex,
        dg: DGGraph,
        u: Union[EntityValueNode, CellNode],
        v: Union[EntityValueNode, CellNode],
        max_n_hop: int = 2,
        bidirection: bool = True,
    ):
        """Find all paths between two nodes in the graph.

        Parameters
        ----------
        qnodes: dictionary of qnodes
        dg: the input data graph
        u: node
        v: node
        max_n_hop: hop start at 1 (current node), n_hop = 2 mean the path can go through an hidden node
        bidirection: set to false telling if we want to just find paths going from u -> v.
            This parameter just for re-using code, as we will call the function changing the order of u & v

        Returns
        -------
        """

        paths = []

        # no qnode in the source
        u_qnodes = (
            [self.wdentities[u.qnode_id]]
            if isinstance(u, EntityValueNode)
            else [self.wdentities[qnode_id] for qnode_id in u.entity_ids]
        )
        if len(u_qnodes) == 0:
            if bidirection:
                return self.kg_path_discovering(
                    kg_object_index,
                    dg,
                    v,
                    u,
                    max_n_hop=max_n_hop,
                    bidirection=False,
                )
            return []

        v_qnodes = (
            [self.wdentities[v.qnode_id]]
            if isinstance(v, EntityValueNode)
            else [self.wdentities[qnode_id] for qnode_id in v.entity_ids]
        )
        if isinstance(v, EntityValueNode):
            # entity in the context
            assert v.context_span is not None
            v_value = v.context_span.get_text_span()
        else:
            v_value = v.value

        for n1 in u_qnodes:
            # count = timer.start("object discovery")
            for n2 in v_qnodes:
                for newpath in self._path_object_search(
                    kg_object_index, n1, n2, max_n_hop=max_n_hop
                ):
                    newpath.sequence.insert(0, DGPathExistingNode(u.id))
                    if isinstance(newpath.sequence[-1], DGPathEdge):
                        newpath.sequence.append(DGPathExistingNode(v.id))
                    paths.append(newpath)
            # count.stop()

            # count = timer.start("literal discovery")
            for newpath in self._path_value_search(n1, self.textparser.parse(v_value)):
                newpath.sequence.insert(0, DGPathExistingNode(u.id))
                if isinstance(newpath.sequence[-1], DGPathEdge):
                    newpath.sequence.append(DGPathExistingNode(v.id))
                paths.append(newpath)
            # count.stop()

        if bidirection:
            paths += self.kg_path_discovering(
                kg_object_index,
                dg,
                v,
                u,
                max_n_hop=max_n_hop,
                bidirection=False,
            )

        return paths

    def _path_value_search(self, source: WDEntity, value):
        matches = []
        for p, stmts in source.props.items():
            if p == "P31":
                # no need to search in the instanceOf property, as the ontology is removed from the databased as they are huge
                continue

            for stmt_i, stmt in enumerate(stmts):
                has_stmt_value = False
                for fn, (match, confidence) in self.literal_match.match(
                    stmt.value, value, skip_unmatch=True
                ):
                    matches.append(
                        DGPath(
                            sequence=[
                                DGPathEdge.p(p),
                                DGPathNodeStatement.from_FromLiteralMatchingFunc(
                                    source.id,
                                    p,
                                    stmt_i,
                                    {"func": fn.__name__, "value": stmt.value},
                                    confidence,
                                ),
                                DGPathEdge.p(p),
                            ]
                        )
                    )
                    has_stmt_value = True

                for q, qvals in stmt.qualifiers.items():
                    for qval in qvals:
                        for fn, (match, confidence) in self.literal_match.match(
                            qval, value, skip_unmatch=True
                        ):
                            if not has_stmt_value:
                                if stmt.value.is_qnode(stmt.value):
                                    pn_stmt_value = DGPathNodeEntity(
                                        stmt.value.as_entity_id()
                                    )
                                else:
                                    pn_stmt_value = DGPathNodeLiteralValue(stmt.value)
                                matches.append(
                                    DGPath(
                                        sequence=[
                                            DGPathEdge.p(p),
                                            DGPathNodeStatement.from_FromWikidataLink(
                                                source.id, p, stmt_i
                                            ),
                                            DGPathEdge.p(p),
                                            pn_stmt_value,
                                        ]
                                    )
                                )
                                has_stmt_value = True

                            matches.append(
                                DGPath(
                                    sequence=[
                                        DGPathEdge.p(p),
                                        DGPathNodeStatement.from_FromLiteralMatchingFunc(
                                            source.id,
                                            p,
                                            stmt_i,
                                            {
                                                "func": fn.__name__,
                                                "value": qval,
                                            },
                                            confidence,
                                        ),
                                        DGPathEdge.q(q),
                                    ]
                                )
                            )
        return matches

    def _path_object_search_v1(
        self,
        kg_object_index: KGObjectIndex,
        source: WDEntity,
        target: WDEntity,
        max_n_hop=2,
    ):
        # max_n_hop = 2 mean we will find a path that go from source to target through an hidden node
        # hop start at 1 (current node)

        if not DGConfigs.ALLOW_SAME_ENT_SEARCH and source.id == target.id:
            return []

        matches = []
        iter = source.props.items()

        for p, stmts in iter:
            if p == "P31":
                # no need to search in the instanceOf property, as the ontology is removed from the databased as they are huge
                continue

            for stmt_i, stmt in enumerate(stmts):
                # add the cell id so that we have different statement for different cell.
                has_stmt_value = False

                # to simplify the design, we do not consider a statement that its value do not exist in KQ
                # due to an error on KG
                if stmt.value.is_qnode(stmt.value):
                    if stmt.value.as_entity_id() not in self.wdentities:
                        # this can happen due to some of the qnodes is in the link, but is missing in the KG
                        # this is very rare so we can employ some check to make sure this is not due to
                        # our wikidata subset
                        is_error_in_kg = any(
                            any(
                                _s.value.is_qnode(_s.value)
                                and _s.value.as_entity_id() in self.wdentities
                                for _s in _stmts
                            )
                            for _p, _stmts in source.props.items()
                        ) or stmt.value.as_entity_id().startswith("L")
                        if not is_error_in_kg:
                            raise Exception(
                                f"Missing qnodes in your KG subset: {stmt.value.as_entity_id()}"
                            )
                        continue

                if stmt.value.is_qnode(stmt.value):
                    # found by match entity
                    if stmt.value.as_entity_id() == target.id:
                        matches.append(
                            DGPath(
                                sequence=[
                                    DGPathEdge.p(p),
                                    DGPathNodeStatement.from_FromWikidataLink(
                                        source.id, p, stmt_i
                                    ),
                                    DGPathEdge.p(p),
                                ]
                            )
                        )
                        has_stmt_value = True
                    elif max_n_hop > 1:
                        stmt_value_qnode_id = stmt.value.as_entity_id()
                        if stmt_value_qnode_id.startswith("L"):
                            assert (
                                stmt_value_qnode_id not in self.wdentities
                            ), "The L nodes (lexical) is not in the Wikidata dump"
                            continue

                        for nextpath in self._path_object_search(
                            kg_object_index,
                            self.wdentities[stmt_value_qnode_id],
                            target,
                            max_n_hop - 1,
                        ):
                            matches.append(
                                DGPath(
                                    sequence=[
                                        DGPathEdge.p(p),
                                        DGPathNodeStatement.from_FromWikidataLink(
                                            source.id, p, stmt_i
                                        ),
                                        DGPathEdge.p(p),
                                        DGPathNodeEntity(stmt_value_qnode_id),
                                    ]
                                    + nextpath.sequence
                                )
                            )
                            has_stmt_value = True

                for q, qvals in stmt.qualifiers.items():
                    for qval in qvals:
                        if qval.is_qnode(qval):
                            if qval.as_entity_id() == target.id:
                                if not has_stmt_value:
                                    if stmt.value.is_qnode(stmt.value):
                                        pn_stmt_value = DGPathNodeEntity(
                                            stmt.value.as_entity_id()
                                        )
                                    else:
                                        pn_stmt_value = DGPathNodeLiteralValue(
                                            stmt.value
                                        )
                                    matches.append(
                                        DGPath(
                                            sequence=[
                                                DGPathEdge.p(p),
                                                DGPathNodeStatement.from_FromWikidataLink(
                                                    source.id, p, stmt_i
                                                ),
                                                DGPathEdge.p(p),
                                                pn_stmt_value,
                                            ]
                                        )
                                    )
                                    has_stmt_value = True

                                matches.append(
                                    DGPath(
                                        sequence=[
                                            DGPathEdge.p(p),
                                            DGPathNodeStatement.from_FromWikidataLink(
                                                source.id, p, stmt_i
                                            ),
                                            DGPathEdge.q(q),
                                        ]
                                    )
                                )
                            elif max_n_hop > 1:
                                qval_qnode_id = qval.as_entity_id()
                                if qval_qnode_id.startswith("L"):
                                    assert (
                                        qval_qnode_id not in self.wdentities
                                    ), "The L nodes (lexical) is not in the Wikidata dump"
                                    continue

                                if qval_qnode_id not in self.wdentities:
                                    # this can happen due to some of the qnodes is in the link, but is missing in the KG
                                    # this is very rare so we can employ some check to make sure this is not due to
                                    # our wikidata subset
                                    is_error_in_kg = any(
                                        any(
                                            _s.value.is_qnode(_s.value)
                                            and _s.value.as_entity_id()
                                            in self.wdentities
                                            for _s in _stmts
                                        )
                                        for _p, _stmts in source.props.items()
                                    ) or qval_qnode_id.startswith("L")
                                    if not is_error_in_kg:
                                        raise Exception(
                                            f"Missing qnodes in your KG subset: {qval_qnode_id}"
                                        )
                                    continue

                                _n_matches = len(matches)
                                for nextpath in self._path_object_search(
                                    kg_object_index,
                                    self.wdentities[qval_qnode_id],
                                    target,
                                    max_n_hop - 1,
                                ):
                                    matches.append(
                                        DGPath(
                                            sequence=[
                                                DGPathEdge.p(p),
                                                DGPathNodeStatement.from_FromWikidataLink(
                                                    source.id, p, stmt_i
                                                ),
                                                DGPathEdge.q(q),
                                                DGPathNodeEntity(qval.as_entity_id()),
                                            ]
                                            + nextpath.sequence
                                        )
                                    )

                                if len(matches) > _n_matches and not has_stmt_value:
                                    if stmt.value.is_qnode(stmt.value):
                                        pn_stmt_value = DGPathNodeEntity(
                                            stmt.value.as_entity_id()
                                        )
                                    else:
                                        pn_stmt_value = DGPathNodeLiteralValue(
                                            stmt.value
                                        )
                                    matches.append(
                                        DGPath(
                                            sequence=[
                                                DGPathEdge.p(p),
                                                DGPathNodeStatement.from_FromWikidataLink(
                                                    source.id, p, stmt_i
                                                ),
                                                DGPathEdge.p(p),
                                                pn_stmt_value,
                                            ]
                                        )
                                    )
                                    has_stmt_value = True

        return matches

    def _path_object_search_v2(
        self,
        kg_object_index: KGObjectIndex,
        source: WDEntity,
        target: WDEntity,
        max_n_hop=2,
    ):
        # max_n_hop = 2 mean we will find a path that go from source to target through an hidden node
        # hop start at 1 (current node)

        if not DGConfigs.ALLOW_SAME_ENT_SEARCH and source.id == target.id:
            return []

        matches = []
        for hop1_path in kg_object_index.iter_hop1_props(source.id, target.id):
            stmt_i = hop1_path.statement_index
            rel = hop1_path.relationship
            stmt = source.props[rel.prop][stmt_i]

            if len(rel.quals) > 0 and not rel.both:
                # the prop doesn't match, have to add it
                if stmt.value.is_qnode(stmt.value):
                    pn_stmt_value = DGPathNodeEntity(stmt.value.as_entity_id())
                else:
                    pn_stmt_value = DGPathNodeLiteralValue(stmt.value)
                matches.append(
                    DGPath(
                        sequence=[
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                source.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                            pn_stmt_value,
                        ]
                    )
                )
            else:
                # the prop match
                matches.append(
                    DGPath(
                        sequence=[
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                source.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                        ]
                    )
                )
            for qual in rel.quals:
                matches.append(
                    DGPath(
                        sequence=[
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                source.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.q(qual),
                        ]
                    )
                )

        if max_n_hop > 1:
            # TODO: need to refactor the current method of searching for paths
            # as we may generate many duplicated segments
            for match_item in kg_object_index.iter_hop2_props(source.id, target.id):
                if match_item.qnode == target.id:
                    continue
                hop1_paths = match_item.hop1
                hop2_paths = match_item.hop2
                middle_qnode = self.wdentities[match_item.qnode]

                # we don't care about transitive here, so we do a cross product
                hop1_seqs = []
                for hop1_path in hop1_paths:
                    rel = hop1_path.relationship
                    stmt_i = hop1_path.statement_index
                    stmt = source.props[rel.prop][stmt_i]

                    if len(rel.quals) > 0 and not rel.both:
                        # we don't have the statement value yet, add it to the matches
                        # the prop doesn't match, have to add it, we don't worry about duplication
                        # as it is resolve during the merge provenance phase
                        if stmt.value.is_qnode(stmt.value):
                            pn_stmt_value = DGPathNodeEntity(stmt.value.as_entity_id())
                        else:
                            pn_stmt_value = DGPathNodeLiteralValue(stmt.value)
                        matches.append(
                            DGPath(
                                sequence=[
                                    DGPathEdge.p(rel.prop),
                                    DGPathNodeStatement.from_FromWikidataLink(
                                        source.id, rel.prop, stmt_i
                                    ),
                                    DGPathEdge.p(rel.prop),
                                    pn_stmt_value,
                                ]
                            )
                        )
                    else:
                        # add prop to the seqs that we need to expand next, and so stmt.value must be a qnode
                        # as it is the middle qnode
                        assert stmt.value.as_entity_id_safe() == middle_qnode.id
                        hop1_seqs.append(
                            [
                                DGPathEdge.p(rel.prop),
                                DGPathNodeStatement.from_FromWikidataLink(
                                    source.id, rel.prop, stmt_i
                                ),
                                DGPathEdge.p(rel.prop),
                                DGPathNodeEntity(middle_qnode.id),
                            ]
                        )

                    for qual in rel.quals:
                        hop1_seqs.append(
                            [
                                DGPathEdge.p(rel.prop),
                                DGPathNodeStatement.from_FromWikidataLink(
                                    source.id, rel.prop, stmt_i
                                ),
                                DGPathEdge.p(qual),
                                DGPathNodeEntity(middle_qnode.id),
                            ]
                        )

                for hop2_path in hop2_paths:
                    rel = hop2_path.relationship
                    stmt_i = hop2_path.statement_index
                    stmt = middle_qnode.props[rel.prop][stmt_i]

                    if len(rel.quals) > 0 and not rel.both:
                        if stmt.value.is_qnode(stmt.value):
                            pn_stmt_value = DGPathNodeEntity(stmt.value.as_entity_id())
                        else:
                            pn_stmt_value = DGPathNodeLiteralValue(stmt.value)

                        hop2_seq = [
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                middle_qnode.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                            pn_stmt_value,
                        ]
                        for hop1_seq in hop1_seqs:
                            matches.append(DGPath(sequence=hop1_seq + hop2_seq))
                    else:
                        hop2_seq = [
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                middle_qnode.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(rel.prop),
                        ]
                        for hop1_seq in hop1_seqs:
                            matches.append(DGPath(sequence=hop1_seq + hop2_seq))

                    for qual in rel.quals:
                        hop2_seq = [
                            DGPathEdge.p(rel.prop),
                            DGPathNodeStatement.from_FromWikidataLink(
                                middle_qnode.id, rel.prop, stmt_i
                            ),
                            DGPathEdge.p(qual),
                        ]
                        for hop1_seq in hop1_seqs:
                            matches.append(DGPath(sequence=hop1_seq + hop2_seq))

        return matches



from typing import Literal


class TypeFeatureConfigs:
    """Holding configuration for type feature extraction"""

    # methods for fixing cycles in KG's ontology classes
    # cycles rarely happen so the recommended method is to fix them manually
    # for example, in Wikidata 2020, we found only 9 cycles out of millions of classes
    # set it to "auto" if you want to automatically fix cycles
    FIX_TYPE_CYCLE: Literal["auto", "manually"] = "manually"



import functools
from collections import defaultdict
from enum import Enum
from operator import xor, attrgetter
from grams.algorithm.data_graph import CellNode
from grams.algorithm.data_graph.dg_graph import DGGraph, EntityValueNode

import networkx as nx
from typing import Dict, Iterable, Mapping, Tuple, Set, List, Optional, Callable, cast

from kgdata.wikidata.models import WDEntity, WDProperty, WDQuantityPropertyStats
from grams.inputs.linked_table import LinkedTable
from grams.algorithm.data_graph import DGNode
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGNode,
    CGStatementNode,
    CGEdge,
)
from grams.algorithm.literal_matchers import TextParser


class QuantityType(Enum):
    Integer = "int"
    Float = "float"


class LinkFeatureExtraction:
    FreqOverRow = "FrequencyOfLinkOverRow"
    FreqOverEntRow = "FrequencyOfLinkOverEntRow"
    FreqOverPossibleLink = "FrequencyOfLinkOverPossibleLink"
    FreqUnmatchOverEntRow = "FrequencyOfUnmatchLinkOverEntRow"
    FreqUnmatchOverPossibleLink = "FrequencyOfUnmatchLinkOverPossibleLink"
    GTE5Link = "GTE5Link"
    GTE10Link = "GTE10Link"
    GTE15Link = "GTE15Link"
    GTE20Link = "GTE20Link"
    GTE30Link = "GTE30Link"
    GTE40Link = "GTE40Link"
    GTE50Link = "GTE50Link"
    NotFuncDep = "NotFuncDep"
    DataTypeMismatch = "DataTypeMismatch"
    HeaderSimilarity = "HeaderSimimilarity"

    def __init__(
        self,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        qnodes: Mapping[str, WDEntity],
        wdprops: Mapping[str, WDProperty],
        wd_num_prop_stats: Mapping[str, WDQuantityPropertyStats],
        sim_fn: Optional[Callable[[str, str], float]] = None,
    ):
        self.table = table
        self.cg = cg
        self.dg = dg
        self.qnodes = qnodes
        self.wdprops = wdprops
        self.wd_num_prop_stats = wd_num_prop_stats
        self.text_parser = TextParser()
        self.sim_fn = sim_fn

        self.cache_get_value_map = {}

    def extract_features(self):
        n_rows = self.table.size()
        freq_over_row = {}
        freq_over_ent_row = {}
        freq_over_pos_link = {}
        freq_unmatch_over_ent_row = {}
        freq_unmatch_over_pos_link = {}
        not_fd_links = {}
        dtype_mismatch_links = {}
        # make none to force setting sim_fn to not None when we want to use the features
        header_sim_links = None if self.sim_fn is None else {}

        gte_link_thresholds = [5, 10, 15, 20, 30, 40, 50]
        gte_link_bins = {k: {} for k in gte_link_thresholds}
        pair2max_possible_link = defaultdict(int)

        # for sid, sdata in self.sg.nodes(data=True):
        for stmt in self.cg.iter_nodes():
            if not isinstance(stmt, CGStatementNode):
                continue

            # in_edge: SGEdge
            # out_edge: SGEdge

            (in_edge,) = self.cg.in_edges(stmt.id)
            unode = self.cg.get_node(in_edge.source)
            # ((uid, _, in_edge),) = list(self.cg.in_edges(stmt.id, data="data"))
            # unode: SGNode = self.cg.nodes[uid]["data"]
            # for _, vid, out_edge in self.cg.out_edges(stmt.id, data="data"):
            for out_edge in self.cg.out_edges(stmt.id):
                # vnode: SGNode = self.cg.nodes[vid]["data"]
                vnode = self.cg.get_node(out_edge.target)
                dg_flows = stmt.get_edges_provenance([out_edge])
                dg_links: Set[Tuple[str, str]] = set()

                sum_prob = 0.0
                for source_flow, target_flows in dg_flows.items():
                    assert source_flow.sg_source_id == unode.id
                    if isinstance(unode, CGColumnNode):
                        # we should only have one target flow as we consider row by row only
                        target_flow, provenances = self._unpack_size1_dict(target_flows)
                        dg_links.add(
                            (source_flow.dg_source_id, target_flow.dg_target_id)
                        )
                        sum_prob += max(provenances, key=attrgetter("prob")).prob
                    else:
                        # this is entity and we may have more than one target flow (to different row)
                        assert isinstance(unode, CGEntityValueNode)
                        for target_flow, provenances in target_flows.items():
                            dg_links.add(
                                (source_flow.dg_source_id, target_flow.dg_target_id)
                            )
                            sum_prob += max(provenances, key=attrgetter("prob")).prob

                # what is the maximum possible links we can have? this ignore the the link so this is used to calculate FreqOverEntRow
                max_possible_ent_rows = (
                    self._get_maximum_possible_ent_links_between_two_nodes(
                        unode.id,
                        vnode.id,
                        self.wdprops[out_edge.predicate].is_data_property(),
                    )
                )

                # calculate the number of links that the source node doesn't have the values as in the target nodes
                # we should consider on what kind of missing. e.g., object missing is more obvious than value mis-match
                # in value mis-match: datetime mis-match is more profound than area/population mis-match
                n_unmatch_links = self._get_n_unmatch_discovered_links(
                    unode.id,
                    vnode.id,
                    in_edge.predicate,
                    out_edge.predicate,
                    dg_links,
                    self.wdprops[out_edge.predicate].is_data_property(),
                )
                n_possible_links = n_unmatch_links + len(dg_links)
                pair2max_possible_link[unode.id, vnode.id] = max(
                    n_possible_links, pair2max_possible_link[unode.id, vnode.id]
                )

                # compute the features
                pair_freq_over_row = sum_prob / n_rows
                pair_freq_over_ent_row = (
                    (sum_prob / max_possible_ent_rows)
                    if max_possible_ent_rows > 0
                    else 0
                )
                # re-calculate it later
                pair_freq_over_pos_link = (sum_prob, (unode.id, vnode.id))
                pair_freq_unmatch_over_ent_row = (
                    (n_unmatch_links / max_possible_ent_rows)
                    if max_possible_ent_rows > 0
                    else 0
                )
                # re-calculate it later
                pair_freq_unmatch_over_pos_link = (
                    n_unmatch_links,
                    (unode.id, vnode.id),
                )
                dtype_mismatch = None
                header_sim = None

                if isinstance(vnode, CGColumnNode):
                    assert out_edge.predicate[0] == "P", "Sanity Check"
                    if in_edge.predicate in self.wd_num_prop_stats and (
                        in_edge.predicate == out_edge.predicate
                        or out_edge.predicate
                        in self.wd_num_prop_stats[in_edge.predicate].qualifiers
                    ):
                        if in_edge.predicate == out_edge.predicate:
                            stat = self.wd_num_prop_stats[in_edge.predicate].value
                        else:
                            stat = self.wd_num_prop_stats[in_edge.predicate].qualifiers[
                                out_edge.predicate
                            ]

                        dtype = self._estimate_col_quantity_type(vnode.column)
                        if dtype is not None and stat.size > 0:
                            # if the property is not quantity, their size is 0
                            is_prop_int = (stat.int_size / stat.size) > 0.99
                            if dtype == QuantityType.Integer:
                                dtype_mismatch = int(not is_prop_int)
                            else:
                                dtype_mismatch = int(is_prop_int)
                    if self.sim_fn is not None:
                        header_sim = self.sim_fn(
                            self.wdprops[out_edge.predicate].label, vnode.label
                        )

                if in_edge.predicate == out_edge.predicate:
                    # copy to the parent property as well
                    key = (unode.id, stmt.id, in_edge.predicate)
                    assert key not in freq_over_row
                    freq_over_row[key] = pair_freq_over_row
                    freq_over_ent_row[key] = pair_freq_over_ent_row
                    freq_over_pos_link[key] = pair_freq_over_pos_link
                    freq_unmatch_over_ent_row[key] = pair_freq_unmatch_over_ent_row
                    freq_unmatch_over_pos_link[key] = pair_freq_unmatch_over_pos_link

                    if isinstance(unode, CGColumnNode) and isinstance(
                        vnode, CGColumnNode
                    ):
                        not_fd_links[key] = int(
                            not self._functional_dependency_test(
                                unode.column, vnode.column
                            )
                        )

                    for threshold, indicator in zip(
                        gte_link_thresholds,
                        self._categorize_num_gte(len(dg_links), gte_link_thresholds),
                    ):
                        gte_link_bins[threshold][key] = indicator

                    if dtype_mismatch is not None:
                        dtype_mismatch_links[key] = dtype_mismatch

                    if header_sim is not None:
                        header_sim_links[key] = header_sim

                key = (stmt.id, vnode.id, out_edge.predicate)
                assert key not in freq_over_row
                freq_over_row[key] = pair_freq_over_row
                freq_over_ent_row[key] = pair_freq_over_ent_row
                freq_over_pos_link[key] = pair_freq_over_pos_link
                freq_unmatch_over_ent_row[key] = pair_freq_unmatch_over_ent_row
                freq_unmatch_over_pos_link[key] = pair_freq_unmatch_over_pos_link

                for threshold, indicator in zip(
                    gte_link_thresholds,
                    self._categorize_num_gte(len(dg_links), gte_link_thresholds),
                ):
                    gte_link_bins[threshold][key] = indicator

                if isinstance(unode, CGColumnNode) and isinstance(vnode, CGColumnNode):
                    not_fd_links[key] = int(
                        not self._functional_dependency_test(unode.column, vnode.column)
                    )

                if dtype_mismatch is not None:
                    dtype_mismatch_links[key] = dtype_mismatch

                if header_sim is not None:
                    header_sim_links[key] = header_sim

        # re-adjust the feature over possible links
        for key, (freq, pair) in freq_over_pos_link.items():
            if pair2max_possible_link[pair] == 0:
                freq_over_pos_link[key] = 0
            else:
                freq_over_pos_link[key] = freq / pair2max_possible_link[pair]
        for key, (freq, pair) in freq_unmatch_over_pos_link.items():
            if pair2max_possible_link[pair] == 0:
                freq_unmatch_over_pos_link[key] = 0
            else:
                freq_unmatch_over_pos_link[key] = freq / pair2max_possible_link[pair]

        return {
            self.FreqOverRow: freq_over_row,
            self.FreqOverEntRow: freq_over_ent_row,
            self.FreqOverPossibleLink: freq_over_pos_link,
            self.FreqUnmatchOverEntRow: freq_unmatch_over_ent_row,
            self.FreqUnmatchOverPossibleLink: freq_unmatch_over_pos_link,
            self.GTE5Link: gte_link_bins[5],
            self.GTE10Link: gte_link_bins[10],
            self.GTE15Link: gte_link_bins[15],
            self.GTE20Link: gte_link_bins[20],
            self.GTE30Link: gte_link_bins[30],
            self.GTE40Link: gte_link_bins[40],
            self.GTE50Link: gte_link_bins[50],
            self.NotFuncDep: not_fd_links,
            self.DataTypeMismatch: dtype_mismatch_links,
            self.HeaderSimilarity: header_sim_links,
        }

    def add_debug_info(self, features: dict, sg=None):
        if sg is None:
            sg = self.cg
        """Add features to edges in the graph for debugging. Features are from the extract_features function"""
        # add the features to edges for debugging purpose
        for feat, feat_data in features.items():
            if feat_data is None:
                continue
            for (uid, vid, eid), val in feat_data.items():
                if sg.has_edge_between_nodes(uid, vid, eid):
                    edge: CGEdge = sg.get_edge_between_nodes(uid, vid, eid)
                    assert feat not in edge.features
                    edge.features[feat] = val

    def _unpack_size1_dict(self, odict: dict):
        """Unpack a dictionary of size 1"""
        lst = list(odict.items())
        assert len(lst) == 1, lst
        return lst[0]

    def _categorize_num_gte(self, num: int, thresholds: List[int]):
        feats = []
        for i in range(len(thresholds)):
            if num >= thresholds[i]:
                feats.append(1)
            else:
                feats.append(0)
        return feats

    def _iter_dg_pair(self, uid: str, vid: str) -> Iterable[Tuple[DGNode, DGNode]]:
        """This function iterate through each pair of data graph nodes between two semantic graph nodes.

        If both sg nodes are entities, we only have one pair.
        If one or all of them are columns, the number of pairs will be the size of the table.
        Otherwise, not support iterating between nodes & statements
        """
        u = self.cg.get_node(uid)
        v = self.cg.get_node(vid)

        if isinstance(u, CGColumnNode) and isinstance(v, CGColumnNode):
            uci = u.column
            vci = v.column
            for ri in range(self.table.size()):
                ucell = self.dg.get_node(f"{ri}-{uci}")
                vcell = self.dg.get_node(f"{ri}-{vci}")
                yield ucell, vcell
        elif isinstance(u, CGColumnNode):
            assert isinstance(v, (CGEntityValueNode, CGLiteralValueNode))
            uci = u.column
            vcell = self.dg.get_node(v.id)
            for ri in range(self.table.size()):
                ucell = self.dg.get_node(f"{ri}-{uci}")
                yield ucell, vcell
        elif isinstance(v, CGColumnNode):
            assert isinstance(u, (CGEntityValueNode, CGLiteralValueNode))
            vci = v.column
            ucell = self.dg.get_node(u.id)
            for ri in range(self.table.size()):
                vcell = self.dg.get_node(f"{ri}-{vci}")
                yield ucell, vcell
        else:
            assert not isinstance(u, CGColumnNode) and not isinstance(v, CGColumnNode)
            yield self.dg.get_node(u.id), self.dg.get_node(v.id)

    def _get_maximum_possible_ent_links_between_two_nodes(
        self, uid: str, vid: str, is_data_predicate: bool
    ):
        """Find the maximum possible links between two nodes (ignore the possible predicates):

        Let M be the maximum possible links we want to find, N is the number of rows in the table.
        1. If two nodes are not columns, M is 1 because it's entity to entity link.
        2. If one node is a column, M = N - U, where U is the number of pairs that cannot have KG discovered links. A
        pair that cannot have KG discovered links is:
            a. If both nodes are columns, and the link is
                * data predicate: the source cell links to no entity.
                * object predicate: the source or target cell link to no entity
            b. If only one node is column, and the link is
                * data predicate:
                    - if the source node must be an entity, then the target must be a column. U is always 0
                    - else then the source node must be is a column and target is a literal value: a cell in the column links to no entity
                * object predicate: a cell in the column links to no entity.
        """
        u = self.cg.get_node(uid)
        v = self.cg.get_node(vid)

        if not isinstance(u, CGColumnNode) and not isinstance(v, CGColumnNode):
            return 1

        # instead of going through each node attach to the node in the semantic graph, we cheat by directly generating
        # the data node ID
        n_rows = self.table.size()
        n_null_entities = 0
        if not (isinstance(u, CGColumnNode) and isinstance(v, CGColumnNode)):
            if is_data_predicate:
                if isinstance(u, (CGEntityValueNode, CGLiteralValueNode)):
                    assert isinstance(u, CGEntityValueNode) and isinstance(
                        v, CGColumnNode
                    )
                    return n_rows

                assert isinstance(u, CGColumnNode) and isinstance(
                    v, (CGEntityValueNode, CGLiteralValueNode)
                )

            if isinstance(u, CGColumnNode):
                ci = u.column
            else:
                assert isinstance(
                    v, CGColumnNode
                ), "Always true -- type system is not smart enough"
                ci = v.column

            for ri in range(n_rows):
                if len(self.dg.get_cell_node(f"{ri}-{ci}").entity_ids) == 0:
                    n_null_entities += 1
        else:
            uci = u.column
            vci = v.column

            for ri in range(n_rows):
                ucell_unk = len(self.dg.get_cell_node(f"{ri}-{uci}").entity_ids) == 0
                vcell_unk = len(self.dg.get_cell_node(f"{ri}-{vci}").entity_ids) == 0

                if is_data_predicate:
                    if ucell_unk:
                        n_null_entities += 1
                elif ucell_unk or vcell_unk:
                    n_null_entities += 1

        return n_rows - n_null_entities

    def _dg_pair_has_possible_ent_links(
        self, dgu: DGNode, dgv: DGNode, is_data_predicate: bool
    ):
        if isinstance(dgu, CellNode) and isinstance(dgv, CellNode):
            # both are cells
            if is_data_predicate:
                # data predicate: source cell must link to some entities to have possible links
                return len(dgu.entity_ids) > 0
            else:
                # object predicate: source cell and target cell must link to some entities to have possible links
                return len(dgu.entity_ids) > 0 and len(dgv.entity_ids) > 0
        elif isinstance(dgu, CellNode):
            # the source is cell, the target will be literal/entity value
            # we have link when source cell link to some entities, doesn't depend on type of predicate
            return len(dgu.entity_ids) > 0
        elif isinstance(dgv, CellNode):
            # the target is cell, the source will be literal/entity value
            if is_data_predicate:
                # data predicate: always has possibe links
                return True
            else:
                # object predicate: have link when the target cell link to some entities
                return len(dgv.entity_ids) > 0
        else:
            # all cells are values, always have link due to how the link is generated in the first place
            return True

    def _get_n_unmatch_discovered_links(
        self,
        uid: str,
        vid: str,
        inpred: str,
        outpred: str,
        uv_links: Set[Tuple[str, str]],
        is_outpred_data_predicate: bool,
    ):
        """Get number of discovered links that don't match due to value differences. This function do not count if:
        * the link between two DG nodes is impossible
        * the property/qualifier do not exist in the WDEntity
        """
        u = self.cg.get_node(uid)
        v = self.cg.get_node(vid)
        is_outpred_qualifier = inpred != outpred

        n_unmatch_links = 0
        for dgu, dgv in self._iter_dg_pair(uid, vid):
            # if has link, then we don't have to count
            if (dgu.id, dgv.id) in uv_links:
                continue

            # ignore pairs that can't have any links
            if not self._dg_pair_has_possible_ent_links(
                dgu, dgv, is_outpred_data_predicate
            ):
                continue

            if isinstance(dgu, CellNode):
                # the source is cell node
                # property doesn't exist in any qnode
                dgu_qnodes = [self.qnodes[qnode_id] for qnode_id in dgu.entity_ids]
                if all(inpred not in qnode.props for qnode in dgu_qnodes):
                    continue

                if is_outpred_qualifier:
                    # qualifier doesn't exist in any qnode
                    has_qual = False
                    for qnode in dgu_qnodes:
                        for stmt in qnode.props.get(inpred, []):
                            if outpred in stmt.qualifiers:
                                has_qual = True
                    if not has_qual:
                        continue

                # # #########################
                # # TODO: remove me, modify on Apr 29, should have a better solution
                # # this is to say like if that properties have multiple values, then it's more likely to miss
                # if all(len(qnode.props.get(inpred, [])) > 2 for qnode in dgu_qnodes):
                #     continue
                # # #########################
            else:
                # the source is entity
                assert isinstance(dgu, EntityValueNode)
                dgu_qnode = self.qnodes[dgu.qnode_id]
                if inpred not in dgu_qnode.props:
                    continue
                if is_outpred_qualifier:
                    if all(
                        outpred not in stmt.qualifiers
                        for stmt in dgu_qnode.props[inpred]
                    ):
                        continue
                # # #########################
                # # TODO: remove me, modify on Apr 29, should have a better solution
                # # this is to say like if that properties have multiple values, then it's more likely to miss
                # if len(dgu_qnode.props.get(inpred, [])) > 2:
                #     continue
                # # #########################
            n_unmatch_links += 1
        return n_unmatch_links

    def _functional_dependency_test(
        self, source_column_index: int, target_column_index: int
    ):
        """Test whether values in the target column is uniquely determiend by the values in the source column. True if
        it's FD.

        Parameters
        ----------
        source_column_index
        target_column_index

        Returns
        -------
        """
        sci = source_column_index
        tci = target_column_index

        # find a mapping from
        source_map = self._get_value_map(source_column_index)
        target_map = {
            ri: key
            for key, rows in self._get_value_map(target_column_index).items()
            for ri in rows
        }

        n_violate_fd = 0
        for key, rows in source_map.items():
            target_keys = {target_map[ri] for ri in rows}
            if len(target_keys) > 1:
                n_violate_fd += 1

        if len(source_map) == 0:
            return True

        if n_violate_fd / len(source_map) > 0.01:
            return False
        return True

    def _estimate_col_quantity_type(self, column_index: int) -> Optional[QuantityType]:
        """These function should be replaced by a ML method"""
        col = self.table.table.get_column_by_index(column_index)
        n_int, n_float = 0, 0
        n_values = 0
        for val in col.values:
            nval = self.text_parser.parse(val)
            if nval.number is not None:
                if isinstance(nval.number, int):
                    n_int += 1
                else:
                    n_float += 1
            if nval.normed_string.lower() not in {"", "na", "n/a"}:
                n_values += 1
        # more than 80% is considered to be the numeric column
        if n_values > 0 and (n_int + n_float) / n_values > 0.95:
            if n_int / (n_int + n_float) > 0.95:
                return QuantityType.Integer
            return QuantityType.Float
        return None

    def _get_value_map(self, column_index: int):
        """Get a map of values in a column to its row numbers (possible duplication).
        This function is not perfect now. The value of the column is considered to be list of entities (if exist) or
        just the value of the cell
        """
        if column_index not in self.cache_get_value_map:
            map = defaultdict(list)
            col = self.table.table.get_column_by_index(column_index)
            for ri in range(self.table.size()):
                links = self.table.links[ri][column_index]
                ents = [link.entity_id for link in links if link.entity_id is not None]
                if len(ents) > 0:
                    key = tuple(ents)
                else:
                    key = col.values[ri].strip()
                map[key].append(ri)
            self.cache_get_value_map[column_index] = dict(map)
        return self.cache_get_value_map[column_index]



import copy
from grams.algorithm.data_graph.dg_graph import (
    DGGraph,
    CellNode,
    EntityValueNode,
    LiteralValueNode,
)
from graph.retworkx import BaseNode, BaseEdge, RetworkXStrDiGraph
import numpy as np
from dataclasses import dataclass
from typing import (
    Dict,
    Optional,
    Union,
    List,
    Set,
    Tuple,
    NamedTuple,
    Any,
)
from uuid import uuid4

import networkx as nx
from kgdata.wikidata.models import WDValue
from grams.algorithm.data_graph import (
    ContextSpan,
    FlowProvenance,
)
import sm.outputs as O
import sm.misc as M


@dataclass
class CGColumnNode(BaseNode[str]):
    id: str
    # column name
    label: str
    # column index
    column: int
    # list nodes' id in the data graph
    nodes: Set[str]

    def clone(self):
        return copy.deepcopy(self)


@dataclass
class CGEntityValueNode(BaseNode[str]):
    id: str
    label: str
    qnode_id: str
    context_span: Optional[ContextSpan]

    @property
    def is_in_context(self):
        return self.context_span is not None

    @property
    def is_statement(self):
        return False

    def clone(self):
        return copy.deepcopy(self)


@dataclass
class CGLiteralValueNode(BaseNode[str]):
    id: str
    label: str
    value: WDValue
    context_span: Optional[ContextSpan]

    @property
    def is_in_context(self):
        return self.context_span is not None

    def clone(self):
        return copy.deepcopy(self)


CGEdgeFlowTarget = NamedTuple(
    "CGEdgeFlowTarget", [("dg_target_id", str), ("sg_target_id", str), ("edge_id", str)]
)
# the source of the flow is not needed but still there for what reason? consistent?
CGEdgeFlowSource = NamedTuple(
    "CGEdgeFlowSource", [("dg_source_id", str), ("sg_source_id", str), ("edge_id", str)]
)


@dataclass
class CGStatementNode(BaseNode[str]):
    id: str
    # (source flow, target_flow) => dg statement id => provenance (we have multiple statement id because
    # for one qnode one property, we may have multiple matches statement
    forward_flow: Dict[CGEdgeFlowSource, Dict[CGEdgeFlowTarget, Set[str]]]
    reversed_flow: Dict[CGEdgeFlowTarget, Dict[CGEdgeFlowSource, Set[str]]]
    flow: Dict[
        Tuple[CGEdgeFlowSource, CGEdgeFlowTarget], Dict[str, List[FlowProvenance]]
    ]

    @staticmethod
    def get_id(source_id: str, source_predicate: str, target_id: str):
        return f"stmt:{source_id}-{source_predicate}-{target_id}"

    @staticmethod
    def new(id: str):
        return CGStatementNode(id, {}, {}, {})

    def clone(self):
        return copy.deepcopy(self)

    def track_provenance(
        self,
        source_flow: CGEdgeFlowSource,
        sid: str,
        target_flow: CGEdgeFlowTarget,
        provenances: List[FlowProvenance],
    ):
        """Track the provenance of which edges goes through the node"""
        if source_flow not in self.forward_flow:
            self.forward_flow[source_flow] = {}
        if target_flow not in self.forward_flow[source_flow]:
            self.forward_flow[source_flow][target_flow] = set()
        self.forward_flow[source_flow][target_flow].add(sid)

        if target_flow not in self.reversed_flow:
            self.reversed_flow[target_flow] = {}
        if source_flow not in self.reversed_flow[target_flow]:
            self.reversed_flow[target_flow][source_flow] = set()
        self.reversed_flow[target_flow][source_flow].add(sid)

        if (source_flow, target_flow) not in self.flow:
            self.flow[source_flow, target_flow] = {}
        assert (
            sid not in self.flow[source_flow, target_flow]
        ), "Gate to make sure the assumption (source_flow, target_flow, dg statement id) is unique correct"
        self.flow[source_flow, target_flow][sid] = provenances

    def compute_freq(
        self,
        _cg: "CGGraph",
        dg: DGGraph,
        edge: "CGEdge",
        is_unique_freq: bool,
    ):
        """Compute the frequency of data edges between"""
        if not is_unique_freq:
            freq = 0
            if edge.target == self.id:
                # from node to statement, we only consider the target that is property of the statement
                # since we always create new statement based on the value
                for (source_flow, target_flow), stmt2prov in self.flow.items():
                    if (
                        source_flow.edge_id != edge.predicate
                        or source_flow.sg_source_id != edge.source
                    ):
                        continue
                    if target_flow.edge_id != edge.predicate:
                        continue
                    freq += 1
            else:
                for (source_flow, target_flow), stmt2prov in self.flow.items():
                    if (
                        target_flow.sg_target_id == edge.target
                        and target_flow.edge_id == edge.predicate
                    ):
                        freq += 1
            return freq

        unique_pairs = set()
        if edge.target == self.id:
            for (source_flow, target_flow), stmt2prov in self.flow.items():
                if (
                    source_flow.sg_source_id != edge.source
                    or source_flow.edge_id != edge.predicate
                ):
                    continue

                # the target is the one contain value of the statement property, which we should have
                # only one
                if target_flow.edge_id != edge.predicate:
                    continue

                dg_source = dg.get_node(source_flow.dg_source_id)
                if isinstance(dg_source, CellNode):
                    source_val = dg_source.value
                else:
                    assert isinstance(dg_source, EntityValueNode)
                    source_val = dg_source.qnode_id

                dg_target = dg.get_node(target_flow.dg_target_id)
                if isinstance(dg_target, CellNode):
                    target_val = dg_target.value
                elif isinstance(dg_target, LiteralValueNode):
                    target_val = dg_target.value.to_string_repr()
                else:
                    assert isinstance(dg_target, EntityValueNode)
                    target_val = dg_target.qnode_id

                unique_pairs.add((source_val, target_val))
        else:
            for source_flow, target_flows in self.forward_flow.items():
                dg_source = dg.get_node(source_flow.dg_source_id)
                if isinstance(dg_source, CellNode):
                    source_val = dg_source.value
                else:
                    assert isinstance(dg_source, EntityValueNode)
                    source_val = dg_source.qnode_id
                for target_flow in target_flows.keys():
                    if (
                        target_flow.sg_target_id == edge.target
                        and target_flow.edge_id == edge.predicate
                    ):
                        dg_target = dg.get_node(target_flow.dg_target_id)
                        if isinstance(dg_target, CellNode):
                            target_val = dg_target.value
                        elif isinstance(dg_target, LiteralValueNode):
                            target_val = dg_target.value.to_string_repr()
                        else:
                            assert isinstance(dg_target, EntityValueNode)
                            target_val = dg_target.qnode_id
                        unique_pairs.add((source_val, target_val))

        return len(unique_pairs)

    def get_edges_provenance(self, edges: List["CGEdge"]):
        """Retrieve all flow provenances that connects these edges.

        Note that in case we have more than one statement per flow, we merge the statements (i.e., merge
        the provenances) to return the best statements (since the provenance store how the link is generated,
        merged the provenance will select the best one).
        """
        select_target_flows = {(e.target, e.predicate) for e in edges}
        links = {}
        for source_flow, target_flows in self.forward_flow.items():
            # source_flow.predicate is always the same, just change to different data node
            # get list of statements that contains all edges
            stmts: Optional[
                Dict[str, Dict[CGEdgeFlowTarget, List[FlowProvenance]]]
            ] = None
            for target_flow in target_flows:
                if (
                    target_flow.sg_target_id,
                    target_flow.edge_id,
                ) not in select_target_flows:
                    continue
                if stmts is None:
                    stmts = {}
                    for stmt_id, provs in self.flow[source_flow, target_flow].items():
                        stmts[stmt_id] = {}
                        stmts[stmt_id][target_flow] = provs
                else:
                    # do intersection because we want to find the stmt contain all edges
                    stmt2provenances = self.flow[source_flow, target_flow]
                    remove_stmt_ids = set(stmts.keys()).difference(
                        stmt2provenances.keys()
                    )
                    for stmt_id in remove_stmt_ids:
                        stmts.pop(stmt_id)
                    for stmt_id, provenances in stmt2provenances.items():
                        if stmt_id not in stmts:
                            continue
                        # no merge since the target flow is unique
                        stmts[stmt_id][target_flow] = provenances

            if stmts is not None and len(stmts) > 0:
                # do a merge to get the provenances (this is the strategy to return the best statement)
                # note that all target flows will be appear, so the first step is to reverse target flow with statement id
                swap_stmts: Dict[CGEdgeFlowTarget, List[FlowProvenance]] = {
                    target_flow: [] for target_flow in next(iter(stmts.values()))
                }
                for stmt_id, _target_flows in stmts.items():
                    for target_flow, provenances in _target_flows.items():
                        swap_stmts[target_flow] = FlowProvenance.merge_lst(
                            swap_stmts[target_flow], provenances
                        )
                links[source_flow] = swap_stmts
        return links


@dataclass
class CGEdge(BaseEdge[str, str]):
    source: str
    target: str
    predicate: str
    features: Dict[str, Any]
    id: int = -1  # set automatically by cg graph

    @property
    def key(self):
        return self.predicate

    def clone(self):
        return copy.deepcopy(self)


CGNode = Union[CGColumnNode, CGEntityValueNode, CGLiteralValueNode, CGStatementNode]
CGEdgeTriple = Tuple[str, str, str]  # source, target, predicate


class CGGraph(RetworkXStrDiGraph[str, CGNode, CGEdge]):
    def get_statement_node(self, uid: str) -> CGStatementNode:
        u = self.get_node(uid)
        assert isinstance(u, CGStatementNode)
        return u

    def remove_dangling_statement(self):
        """Remove statement nodes that have no incoming nodes or no outgoing edges"""
        for s in self.nodes():
            if isinstance(s, CGStatementNode) and (
                self.in_degree(s.id) == 0 or self.out_degree(s.id) == 0
            ):
                self.remove_node(s.id)

    def remove_standalone_nodes(self):
        """Remove nodes that do not any incoming / outgoing edges"""
        for u in self.nodes():
            if self.degree(u.id) == 0:
                self.remove_node(u.id)

    @staticmethod
    def graphviz_props():
        """Return necessary props for the function graph.viz.graphviz.draw"""

        def node_label(u: CGNode):
            if isinstance(u, CGColumnNode):
                return f"{u.label} (column {u.column})"
            if isinstance(u, CGStatementNode):
                return u.id
            return u.label

        def node_styles(u: CGNode):
            if isinstance(u, CGColumnNode):
                return dict(
                    shape="plaintext",
                    style="filled",
                    fillcolor="gold",
                )
            return dict(
                shape="ellipse",
                style="filled",
                color="white",
                fillcolor="lightgray",
            )

        return {"node_label": node_label, "node_styles": node_styles}



from __future__ import annotations
import re, os, uuid, logging
import shutil
from typing import (
    Any,
    Dict,
    Generic,
    List,
    MutableMapping,
    Optional,
    Set,
    TypeVar,
)
from datetime import datetime
from numpy import ndarray
from pslpython.partition import Partition
from pslpython.predicate import Predicate
from pslpython.rule import Rule
from pslpython.model import Model, ModelError
from dataclasses import dataclass
from loguru import logger
from pathlib import Path
from copy import copy
from functools import partial
from operator import itemgetter


K = TypeVar("K")


class PSLModel:
    def __init__(
        self,
        predicates: List[Predicate],
        rules: RuleContainer,
        temp_dir: Optional[str] = None,
        ignore_predicates_not_in_rules: bool = False,
        required_predicates: Optional[Set[str]] = None,
    ):
        self.model = Model("psl-model")
        self.rules = rules
        self.name2predicate: Dict[str, Predicate] = {}

        if temp_dir is not None:
            self.temp_dir = temp_dir
        else:
            Path("/tmp/pslpython").mkdir(exist_ok=True, parents=True)
            self.temp_dir = f"/tmp/pslpython/r{datetime.now().strftime('%y%m%d-%H%M%S')}-{str(uuid.uuid4()).replace('-', '')}"
        assert not Path(self.temp_dir).exists(), f"{self.temp_dir} already exists"

        if ignore_predicates_not_in_rules:
            if required_predicates is None:
                required_predicates = set()
            using_predicates = []
            for p in predicates:
                if p.name() in required_predicates:
                    using_predicates.append(p)
                    continue

                pname = p.name() + "("
                subrules = [
                    r
                    for r in rules.values()
                    if r._rule_body is not None and r._rule_body.find(pname) != -1
                ]
                ignore = True

                if len(subrules) > 0:
                    for r in subrules:
                        body: str = r._rule_body  # type: ignore
                        start = 0
                        while (idx := body.find(pname, start)) != -1:
                            if idx == 0:
                                ignore = False
                                break
                            if re.match("[a-zA-Z0-9_]", body[idx - 1]) is None:
                                ignore = False
                                break
                            start = idx + 1
                        if not ignore:
                            break

                if not ignore:
                    using_predicates.append(p)
                else:
                    logger.debug(
                        "Ignoring predicate {} as it can't be found in any rule",
                        p.name(),
                    )
            predicates = using_predicates

        for p in predicates:
            self.model.add_predicate(p)
            self.name2predicate[p.name()] = p

        for r in rules.values():
            self.model.add_rule(r)

        # setup custom logger
        self.logs = []
        self.logger = logging.getLogger(name="psl-model")
        self.logger.addHandler(InMemHandler(self.logs))

    def parameters(self) -> Dict[str, float]:
        """Return parameters of the model"""
        params = {}
        for name, rule in self.rules.items():
            params[name] = rule.weight()
        return params

    def set_parameters(self, params: Dict[str, float]):
        """Set parameters of the model"""
        for name, rule in self.rules.items():
            if rule.weighted():
                rule.set_weight(params[name])

    def extend_rules(self, rules: RuleContainer, predicates: List[Predicate]):
        """Extend the model with new rules and predicates. This is usually used
        to add rules that were pre-grounded.

        Note: this function must be called before `set_data` or `predict`.
        """
        self.rules.update(rules)
        for r in rules.values():
            self.model.add_rule(r)

        for p in predicates:
            if p.name() not in self.name2predicate:
                self.model.add_predicate(p)
                self.name2predicate[p.name()] = p

    def set_data(
        self,
        observations: Dict[str, list],
        targets: Dict[str, list],
        truth: Dict[str, list],
        force_setall: bool = True,
    ):
        """Set data of the current model"""
        pnames = set(observations.keys()).union(targets, truth)

        for pname in pnames:
            p = self.name2predicate[pname]
            p.clear_data()

            for partition, partition_data in [
                (Partition.OBSERVATIONS, observations),
                (Partition.TARGETS, targets),
                (Partition.TRUTH, truth),
            ]:
                if pname in partition_data and len(partition_data[pname]) > 0:
                    p.add_data(partition, partition_data[pname])

        if force_setall and len(self.name2predicate) != len(pnames):
            raise Exception(
                "Missing predicates: "
                + ", ".join(list(set(self.name2predicate.keys()).difference(pnames)))
            )

    def predict(
        self,
        observations: Dict[str, list],
        targets: Dict[str, list],
        truth: Optional[Dict[str, list]] = None,
        force_setall: bool = True,
        cleanup_tempdir: bool = True,
        retry: bool = True,
    ):
        self.set_data(observations, targets, truth or {}, force_setall)

        for _ in range(3):
            try:
                resp = self.model.infer(
                    logger=self.logger,
                    additional_cli_optons=[
                        "--h2path",
                        os.path.join(self.temp_dir, "h2"),
                    ],
                    temp_dir=self.temp_dir,
                    cleanup_temp=False,
                )
                break
            except ModelError as e:
                # error related to parser (not jvm), no retry
                # non_recoverable_error = any(
                #     line.find(k) != -1
                #     for line in self.logs
                #     for k in [
                #         "org.linqs.psl.parser.antlr.PSLParser",
                #         "Unique index or primary key violation",
                #     ]
                # )
                non_recoverable_error = False
                self.log_errors(e)
                if not retry or non_recoverable_error:
                    raise

            logger.info("Retry PSL inference...")
        else:
            raise ModelError("PSL inference failed")

        if cleanup_tempdir:
            shutil.rmtree(self.temp_dir)

        output = {}
        for p, df in resp.items():
            psize = len(p._types)
            output[p.name()] = {
                tuple((r[i] for i in range(psize))): r["truth"]
                for _, r in df.iterrows()
            }
        return output

    def debug(self, idmap: Optional[IDMap] = None):
        """Write data of predicates into a file for debugging"""
        outdir = Path(self.temp_dir) / "debug"
        outdir.mkdir(exist_ok=True, parents=True)

        def remap(row: ndarray, size: int, idmap: IDMap):
            row = copy(row)
            for i in range(size):
                row[i] = idmap.im(row[i])
            return row

        for pname, p in self.name2predicate.items():
            psize = len(p._types)
            for partition, df in p.data().items():
                if df.size == 0:
                    continue

                if idmap is not None:
                    df = df.apply(
                        partial(remap, size=psize, idmap=idmap), axis=1, raw=True
                    )
                df.to_csv(outdir / f"{pname}.{partition}.csv", index=False)

    def log_errors(self, error: Exception, clear: bool = True):
        logger.error(str(error))
        print("\n".join(self.logs))
        if clear:
            self.logs = []

    @staticmethod
    def normalize_probs(
        target_values: Dict[K, float], eps: float = 0.001, threshold: float = 0.0
    ) -> Dict[K, float]:
        """The predicted probabilities can be noisy due to numerical optimizations, i.e., equal variables
        can have slightly different scores. This function groups values that are close
        within the range [-eps, +eps] together, and replace them with the average value
        to reduce the noises.

        Args:
            target_values: probability of grounded values of a predicate
            eps: group values within [-eps, +eps] together
            threshold: remove values with probability less than threshold

        Return:
            A dictionary of normalized probabilities >= threshold
        """
        if eps == 0.0 and threshold == 0.0:
            return target_values

        norm_probs = {}
        lst = sorted(
            [x for x in target_values.items() if x[1] >= threshold], key=itemgetter(1)
        )

        if len(lst) == 0:
            return {}

        clusters = []
        pivot = 1
        clusters = [[lst[0]]]
        while pivot < len(lst):
            x = lst[pivot - 1][1]
            y = lst[pivot][1]
            if (y - x) <= eps:
                # same clusters
                clusters[-1].append(lst[pivot])
            else:
                # different clusters
                clusters.append([lst[pivot]])
            pivot += 1
        for cluster in clusters:
            avg_prob = sum([x[1] for x in cluster]) / len(cluster)
            for k, _prob in cluster:
                norm_probs[k] = avg_prob
        return norm_probs


class InMemHandler(logging.Handler):
    def __init__(self, storage: list):
        logging.Handler.__init__(self)
        self.storage = storage

    def emit(self, record):
        self.storage.append(self.format(record))


class IDMap(Generic[K]):
    def __init__(self, counter: int = 0):
        self.counter = counter
        self.map: Dict[K, str] = {}
        self.invert_map: Dict[str, K] = {}

    def m(self, key: K) -> str:
        """Get a new key from old key"""
        if key not in self.map:
            new_key = f"i-{self.counter}"
            self.map[key] = new_key
            self.invert_map[new_key] = key
            self.counter += 1
        return self.map[key]

    def im(self, new_key: str) -> K:
        """Get the old key from the new key"""
        return self.invert_map[new_key]

    def to_readable(self) -> dict[str, str]:
        """Transform the IDMap so that the keys reserve the original values as much as possible

        Returns:
            dict: A dictionary mapping the original keys to the new keys so that you can
                  update objects that has been mapped before
        """
        map, invert_map = {}, {}
        trans = {}
        for v, k in self.map.items():
            newk = "n-" + re.sub(r"""[:"'{}+=]""", "-", str(v))
            assert newk not in map
            map[v] = newk
            invert_map[newk] = v
            trans[k] = newk

        self.map = map
        self.invert_map = invert_map

        return trans


class ReadableIDMap(IDMap[K]):
    def __init__(self):
        self.map: Dict[K, str] = {}
        self.invert_map: Dict[str, K] = {}

    def m(self, key: K) -> str:
        """Get a new key from old key"""
        if key not in self.map:
            new_key = "n-" + re.sub(r"""[:"'{}+=]""", "-", str(key))
            self.map[key] = new_key
            self.invert_map[new_key] = key
        return self.map[key]

    def im(self, new_key: str) -> K:
        """Get the old key from the new key"""
        return self.invert_map[new_key]


class RuleContainer(MutableMapping[str, Rule]):
    def __init__(self, rules: Optional[Dict[str, Rule]] = None):
        self.rules: Dict[str, Rule] = rules if rules is not None else {}

    def __len__(self):
        return len(self.rules)

    def __iter__(self):
        return self.rules.__iter__()

    def items(self):
        return self.rules.items()

    def keys(self):
        return self.rules.keys()

    def values(self):
        return self.rules.values()

    def __contains__(self, key: str) -> bool:
        return key in self.rules

    def __setitem__(self, key: str, value: Rule):
        assert key not in self.rules, key
        self.rules[key] = value

    def __getitem__(self, key: str) -> Rule:
        return self.rules[key]

    def __delitem__(self, key: str) -> None:
        self.rules.pop(key)

    def print(self):
        """Print list of rules"""
        print("=" * 10, "Rules")
        for name, rule in self.rules.items():
            print(f"{name}: {rule._rule_body}")
        print("=" * 10)



from __future__ import annotations
import copy
from dataclasses import dataclass
from operator import itemgetter
from urllib.parse import urlparse, unquote_plus
from grams.inputs.context import ContentHierarchy
from kgdata.wikipedia.misc import get_title_from_url

import requests, re, pandas as pd
from bs4 import BeautifulSoup
from bs4.element import Tag, NavigableString
from typing import Callable, List, Dict, Mapping, Optional

from sm.prelude import I, M
import grams.inputs as GI
from grams.config import DATA_DIR
from grams.html_table_parser.parsing_exception import (
    InvalidCellSpanException,
    InvalidColumnSpanException,
    OverlapSpanException,
)


@dataclass
class HTMLTableCellHTMLElement:
    # html tag (lower case)
    tag: str
    start: int
    # end (exclusive)
    end: int
    # html attributes
    attrs: Dict[str, str]
    children: List["HTMLTableCellHTMLElement"]

    def post_order(self):
        for c in self.children:
            for el in c.post_order():
                yield el
        yield self

    def clone(self):
        return HTMLTableCellHTMLElement(
            self.tag,
            self.start,
            self.end,
            copy.copy(self.attrs),
            [c.clone() for c in self.children],
        )


@dataclass
class HTMLTableCell:
    # text value of the cell
    value: str
    rowspan: int
    colspan: int
    # html of the cell
    html: str
    # list of html elements that created this cell
    # except that:
    # - BR and HR are replaced by `\n` character and not in this list
    # - div, span are container and won't in the list
    # for more details, look at the HTMLTableParser._extract_cell_recur function
    elements: List[HTMLTableCellHTMLElement]

    # original row & col span
    original_rowspan: Optional[int] = None
    original_colspan: Optional[int] = None

    def travel_elements_post_order(self):
        for el in self.elements:
            for pointer in el.post_order():
                yield pointer

    def clone(self):
        return HTMLTableCell(
            value=self.value,
            rowspan=self.rowspan,
            colspan=self.colspan,
            html=self.html,
            elements=[el.clone() for el in self.elements],
            original_rowspan=self.original_rowspan,
            original_colspan=self.original_colspan,
        )


@dataclass
class HTMLTableRow:
    cells: List[HTMLTableCell]
    # html attributes of the tr elements
    attrs: Dict[str, str]


@dataclass
class HTMLTable:
    page_url: str
    # value of html caption
    caption: str
    # html attributes of the table html element
    attrs: Dict[str, str]
    # context
    context: List[ContentHierarchy]
    # list of rows in the table
    rows: List[HTMLTableRow]

    def span(self) -> HTMLTable:
        """Span the table by copying values to merged field"""
        if len(self.rows) == 0:
            return self

        pi = 0
        data = []
        pending_ops = {}

        # >>> begin find the max #cols
        # calculate the number of columns as some people may actually set unrealistic colspan as they are lazy..
        # I try to make its behaviour as much closer to the browser as possible.
        # one thing I notice that to find the correct value of colspan, they takes into account the #cells of rows below the current row
        # so we may have to iterate several times
        cols = [0 for _ in range(len(self.rows))]
        for i, row in enumerate(self.rows):
            cols[i] += len(row.cells)
            for cell in row.cells:
                if cell.rowspan > 1:
                    for j in range(1, cell.rowspan):
                        if i + j < len(cols):
                            cols[i + j] += 1

        _row_index, max_ncols = max(enumerate(cols), key=itemgetter(1))
        # sometimes they do show an extra cell for over-colspan row, but it's not consistent or at least not easy for me to find the rule
        # so I decide to not handle that. Hope that we don't have many tables like that.
        # >>> finish find the max #cols

        for row in self.rows:
            new_row = []
            pj = 0
            for cell_index, cell in enumerate(row.cells):
                cell = cell.clone()
                cell.original_colspan = cell.colspan
                cell.original_rowspan = cell.rowspan
                cell.colspan = 1
                cell.rowspan = 1

                # adding cell from the top
                while (pi, pj) in pending_ops:
                    new_row.append(pending_ops[pi, pj].clone())
                    pending_ops.pop((pi, pj))
                    pj += 1

                # now add cell and expand the column
                for _ in range(cell.original_colspan):
                    if (pi, pj) in pending_ops:
                        # exception, overlapping between colspan and rowspan
                        raise OverlapSpanException()
                    new_row.append(cell.clone())
                    for ioffset in range(1, cell.original_rowspan):
                        # no need for this if below
                        # if (pi+ioffset, pj) in pending_ops:
                        #     raise OverlapSpanException()
                        pending_ops[pi + ioffset, pj] = cell
                    pj += 1

                    if pj >= max_ncols:
                        # our algorithm cannot handle the case where people are bullying the colspan system, and only can handle the case
                        # where the span that goes beyond the maximum number of columns is in the last column.
                        if cell_index != len(row.cells) - 1:
                            raise InvalidColumnSpanException()
                        else:
                            break

            # add more cells from the top since we reach the end
            while (pi, pj) in pending_ops and pj < max_ncols:
                new_row.append(pending_ops[pi, pj].clone())
                pending_ops.pop((pi, pj))
                pj += 1

            data.append(HTMLTableRow(cells=new_row, attrs=copy.copy(row.attrs)))
            pi += 1

        # len(pending_ops) may > 0, but fortunately, it doesn't matter as the browser also does not render that extra empty lines
        return HTMLTable(
            page_url=self.page_url,
            caption=self.caption,
            attrs=copy.copy(self.attrs),
            context=[copy.deepcopy(c) for c in self.context],
            rows=data,
        )

    def pad(self) -> HTMLTable:
        """Pad the irregular table (missing cells) to make it become regular table.

        This function only return new table when it's padded
        """
        if len(self.rows) == 0:
            return self

        ncols = len(self.rows[0].cells)
        is_regular_table = all(len(r.cells) == ncols for r in self.rows)
        if is_regular_table:
            return self

        max_ncols = max(len(r.cells) for r in self.rows)
        default_cell = HTMLTableCell(
            value="",
            rowspan=1,
            colspan=1,
            html="",
            elements=[],
            original_rowspan=1,
            original_colspan=1,
        )

        rows = []
        for r in self.rows:
            row = HTMLTableRow(
                cells=[c.clone() for c in r.cells], attrs=copy.copy(r.attrs)
            )
            while len(row.cells) < max_ncols:
                row.cells.append(default_cell.clone())
            rows.append(row)

        return HTMLTable(
            page_url=self.page_url,
            caption=self.caption,
            attrs=copy.copy(self.attrs),
            context=[copy.deepcopy(c) for c in self.context],
            rows=rows,
        )

    def as_df(self):
        return pd.DataFrame([[c.value for c in r.cells] for r in self.rows])

    def as_relational_linked_table(
        self,
        table_id: Optional[str] = None,
        get_entity_from_url: Optional[Callable[[str], Optional[str]]] = None,
        get_entity_from_title: Optional[Callable[[str], Optional[str]]] = None,
    ):
        assert len(self.rows) > 0
        header = [c.value for c in self.rows[0].cells]
        table = I.ColumnBasedTable(
            table_id or self.page_url,
            [
                I.Column(
                    ci,
                    cname,
                    [self.rows[ri].cells[ci].value for ri in range(1, len(self.rows))],
                )
                for ci, cname in enumerate(header)
            ],
        )

        if get_entity_from_url is None:
            if get_entity_from_title is not None:
                get_entity_from_url = lambda url: get_entity_from_title(
                    get_title_from_url(url)
                )
            else:
                get_entity_from_url = lambda _: None

        links = []
        for ri, row in enumerate(self.rows[1:]):
            rlinks = []
            for ci in range(len(header)):
                clinks = []
                for el in row.cells[ci].travel_elements_post_order():
                    if el.tag == "a":
                        entity_id = get_entity_from_url(el.attrs["href"])
                        if entity_id is not None:
                            candidates = [GI.CandidateEntity(entity_id, 1.0)]
                        else:
                            candidates = []
                        clinks.append(
                            GI.Link(
                                el.start,
                                el.end,
                                el.attrs["href"],
                                entity_id,
                                candidates,
                            )
                        )
                rlinks.append(clinks)
            links.append(rlinks)
        return GI.LinkedTable(
            table,
            GI.Context(
                page_url=self.page_url,
                page_title=get_title_from_url(self.page_url),
                page_entity_id=get_entity_from_url(self.page_url),
                content_hierarchy=self.context,
            ),
            links,
        )



from dataclasses import dataclass
from operator import attrgetter
from typing import Dict, Iterator, Mapping, Set, List, Optional, Callable, Type, TypeVar


@dataclass
class Tree:
    @dataclass
    class HierarchyRecord:
        id: str
        duplicated: bool
        depth: int

    id: str
    depth: int
    children: List["Tree"]
    score: Optional[float] = None

    def get_flatten_hierarchy(self, dedup: bool = False) -> List[HierarchyRecord]:
        """Flatten the tree into a flatten hierarchy"""
        stack = [self]
        output = []
        seen_ids = set()
        while len(stack) > 0:
            node = stack.pop()
            dup = node.id in seen_ids
            output.append(Tree.HierarchyRecord(node.id, dup, node.depth))
            seen_ids.add(node.id)

            if not dedup or not dup:
                # do not travel the children
                for c in reversed(node.children):
                    stack.append(c)
        return output

    def sort(self, key=None, reverse: bool = False):
        self.children.sort(key=key or attrgetter("score"), reverse=reverse)
        for c in self.children:
            c.sort(key, reverse)

    def adjust_depth(self, depth: int):
        self.depth = depth
        for c in self.children:
            c.adjust_depth(depth + 1)
        return self

    def update_score(self, score_fn: Callable[["Tree"], float]):
        """Use to adjust score of a node in the tree based on its children. The score is going to be used for sorting"""
        for c in self.children:
            c.update_score(score_fn)
        self.score = score_fn(self)
        return self

    def preorder(
        self,
        fn: Callable[["Tree", List["Tree"]], None],
        path: Optional[List["Tree"]] = None,
    ):
        if path is None:
            path = []
        fn(self, path)
        path.append(self)
        for child in self.children:
            child.preorder(fn, path)
        path.pop()

    def clone(self):
        return Tree(self.id, self.depth, [c.clone() for c in self.children], self.score)


@dataclass
class Forest:
    trees: List[Tree]

    def get_flatten_hierarchy(self, dedup: bool = False) -> List[Tree.HierarchyRecord]:
        return Tree(None, -1, self.trees).get_flatten_hierarchy(dedup)[1:]

    def sort(self, key=None, reverse: bool = False):
        for tree in self.trees:
            tree.sort(key, reverse)
        self.trees.sort(key=key or attrgetter("score"), reverse=reverse)
        return self

    def update_score(self, score_fn: Callable[["Tree"], float]):
        for tree in self.trees:
            tree.update_score(score_fn)
        return self

    def preorder(self, fn: Callable[["Tree"], None]):
        for tree in self.trees:
            tree.preorder(fn)


K = TypeVar("K")
IV = TypeVar("IV")
V = TypeVar("V")


class IndirectDictAccess(Mapping[K, V]):
    def __init__(self, odict: Mapping[K, IV], access: Callable[[IV], V]):
        self.odict = odict
        self.access = access

    def __getitem__(self, item: K) -> V:
        return self.access(self.odict[item])

    def __contains__(self, item: K) -> bool:
        return item in self.odict

    def __iter__(self) -> Iterator[K]:
        return self.odict.__iter__()

    def __len__(self) -> int:
        return self.odict.__len__()


def reorder2tree(lst: List[str], super_relationships: Dict[str, Set[str]]) -> Forest:
    """This function is very forgiving as it doesn't throw exception
    when the qnode is not in the super_relationship dictionary
    """
    if len(lst) == 0:
        return Forest([])

    roots = []
    graph: Dict[str, Tree] = {}

    # unknown class; we do have this case in some dumps (tag to wrong qnodes)
    lst = [u for u in lst if u in super_relationships]

    for i, u in enumerate(lst):
        parents = set()
        for j, v in enumerate(lst):
            if i == j:
                continue
            if v in super_relationships[u]:
                parents.add(v)

        if len(parents) > 1:
            # remove grand parents
            dparents = set()
            for v1 in parents:
                if all(v1 == v2 or v1 not in super_relationships[v2] for v2 in parents):
                    dparents.add(v1)
            parents = dparents

        if len(parents) == 0:
            roots.append(u)

        if u not in graph:
            graph[u] = Tree(u, -1, [])
        for v in parents:
            if v not in graph:
                graph[v] = Tree(v, -1, [])
            graph[v].children.append(graph[u])

    if len(roots) == 0:
        # we have cycle.. try to break it, or we can just throw exception
        # this is going to be very rare based on previous analysis (6 cycles found)
        raise Exception("Found cycles")

    # want to make sure that we have is not a graph
    visited = set()

    def is_acyclic(uid):
        if uid in visited:
            return False
        visited.add(uid)
        for c in graph[uid].children:
            if not is_acyclic(c.id):
                return False
        visited.remove(uid)
        return True

    for root in roots:
        if not is_acyclic(root):
            raise Exception("Found cycles")

    # clone so that we can get correct depth
    return Forest([graph[root].clone().adjust_depth(0) for root in roots])



from collections import defaultdict
from typing import Callable, Dict, List, Mapping, Optional, Tuple
from grams.algorithm.data_graph.dg_graph import DGGraph
from kgdata.wikidata.deprecated.wdclass import WDClass

import networkx as nx
import numpy as np

from grams.algorithm.data_graph import CellNode
from grams.algorithm.literal_matchers import TextParser
from grams.algorithm.candidate_graph.cg_graph import CGColumnNode, CGGraph
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import WDEntity, WDProperty, WDQuantityPropertyStats


class TypeFeatureExtraction:
    Freq = "TypeFreq"
    FreqOverRow = "TypeFreqOverRow"
    FreqInheritOverRow = "TypeFreqInheritOverRow"
    HeaderSimilarity = "TypeHeaderSimilarity"

    def __init__(
        self,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        qnodes: Mapping[str, WDEntity],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
        wd_num_prop_stats: Mapping[str, WDQuantityPropertyStats],
        sim_fn: Optional[Callable[[str, str], float]] = None,
    ):
        self.table = table
        self.cg = cg
        self.dg = dg
        self.qnodes = qnodes
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.wd_num_prop_stats = wd_num_prop_stats
        self.text_parser = TextParser()
        self.sim_fn = sim_fn

        # self.transitive_props = [p.id for p in self.wdprops.values() if p.is_transitive()]
        self.hierarchy_props = {"P131", "P276"}

    def extract_features(self):
        freq_type = {}
        freq_over_row = {}
        freq_inherit_over_row = {}
        cell2qnodes: Dict[str, List[Tuple[WDEntity, float]]] = {}
        column2types = {}
        header_sim_types = {}

        # for uid, udata in self.cg.nodes(data=True):
        for u in self.cg.iter_nodes():
            # u: SGColumnNode = udata["data"]
            if not isinstance(u, CGColumnNode):
                continue

            # cells in this column
            cells: List[CellNode] = [self.dg.get_cell_node(cid) for cid in u.nodes]
            covered_fractions = [
                sum(
                    span.length
                    for spans in cell.entity_spans.values()
                    for span in spans
                )
                / max(len(cell.value), 1)
                for cell in cells
                if len(cell.entity_ids) > 0
            ]
            if len(covered_fractions) == 0:
                continue
            avg_covered_fractions = np.mean(covered_fractions)
            if avg_covered_fractions < 0.8:
                continue

            for cell in cells:
                self.add_merge_qnodes(cell, cell2qnodes)

            # calculate type to frequency
            type2freq = defaultdict(float)
            inherit_type2freq = defaultdict(float)

            for cell in cells:
                classes = {}
                for qnode, prob in cell2qnodes[cell.id]:
                    for stmt in qnode.props.get("P31", []):
                        classes[stmt.value.as_entity_id()] = max(
                            prob, classes.get(stmt.value.as_entity_id(), 0)
                        )
                all_classes = self.retrieve_parents(classes, 1)
                for c, prob in classes.items():
                    type2freq[c] += prob
                for c, prob in all_classes.items():
                    inherit_type2freq[c] += prob

            for c, freq in type2freq.items():
                freq_type[u.id, c] = freq
                freq_over_row[u.id, c] = freq / self.table.size()
            for c, freq in inherit_type2freq.items():
                freq_inherit_over_row[u.id, c] = freq / self.table.size()
            column2types[u.id] = list(inherit_type2freq.keys())
            if len(u.label) > 0 and self.sim_fn is not None:
                for c, freq in inherit_type2freq.items():
                    header_sim_types[u.id, c] = self.sim_fn(
                        u.label, self.wdclasses[c].label
                    )
        return {
            self.Freq: freq_type,
            self.FreqOverRow: freq_over_row,
            self.FreqInheritOverRow: freq_inherit_over_row,
            self.HeaderSimilarity: header_sim_types,
            "_column_to_types": column2types,
        }

    def retrieve_parents(self, classes: Dict[str, float], distance: int):
        """Get parents of classes within distance from the node.
        Distance 1 is the parent, Distance 2 is the grandparent.
        Return the classes and their parents
        """
        if distance <= 0:
            return classes
        output = classes.copy()
        for klass, prob in classes.items():
            for parent_klass in self.wdclasses[klass].parents:
                output[parent_klass] = max(output.get(parent_klass, 0), prob)
        if distance > 1:
            return self.retrieve_parents(output, distance - 1)
        return output

    def add_merge_qnodes(
        self, cell: CellNode, cell2qnodes: Dict[str, List[Tuple[WDEntity, float]]]
    ):
        # merge qnodes that are sub of each other
        # attempt to merge qnodes (spatial) if they are contained in each other
        # we should go even higher order
        assert cell.id not in cell2qnodes

        if len(cell.entity_ids) > 1:
            # attempt to merge qnodes (spatial) if they are contained in each other
            # we should go even higher order
            ignore_qnodes = set()
            for q0_id in cell.entity_ids:
                q0 = self.qnodes[q0_id]
                vals = {
                    stmt.value.as_entity_id()
                    for p in self.hierarchy_props
                    for stmt in q0.props.get(p, [])
                }
                for q1_id in cell.entity_ids:
                    if q0_id == q1_id:
                        continue
                    if q1_id in vals:
                        # q0 is inside q1, ignore q1
                        ignore_qnodes.add(q1_id)
            qnode_lst = [
                self.qnodes[q_id]
                for q_id in cell.entity_ids
                if q_id not in ignore_qnodes
            ]
        elif len(cell.entity_ids) > 0:
            qnode_lst = [self.qnodes[cell.entity_ids[0]]]
        else:
            qnode_lst = []

        qnode2prob = {}
        for link in self.table.links[cell.row][cell.column]:
            for c in link.candidates:
                qnode2prob[c.entity_id] = max(
                    qnode2prob.get(c.entity_id, 0), c.probability
                )

        cell2qnodes[cell.id] = [(qnode, qnode2prob[qnode.id]) for qnode in qnode_lst]



from typing import Dict, List, Mapping
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEdge,
    CGEdgeFlowSource,
    CGEdgeFlowTarget,
    CGEntityValueNode,
    CGGraph,
    CGLiteralValueNode,
    CGStatementNode,
)
from grams.algorithm.data_graph.dg_graph import (
    CellNode,
    DGGraph,
    DGNode,
    EdgeFlowSource,
    EdgeFlowTarget,
    EntityValueNode,
    LiteralValueNode,
    StatementNode,
)
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import WDEntity, WDClass, WDProperty
from sm.prelude import M


class CGFactory:
    def __init__(
        self,
        wdentities: Mapping[str, WDEntity],
        wdentity_labels: Mapping[str, str],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
    ):
        self.wdentities = wdentities
        self.wdentity_labels = wdentity_labels
        self.wdclasses = wdclasses
        self.wdprops = wdprops

    def create_cg(self, table: LinkedTable, dg: DGGraph) -> CGGraph:
        """Create candidate graph from data graph"""
        cg = CGGraph()
        # first step: add node to the graph
        # for uid, udata in dg.iter_nodes():
        #     u: DGNode = udata["data"]
        for u in dg.iter_nodes():
            if isinstance(u, StatementNode):
                # we can have more than one predicates per entity column, these are represented in the statement
                continue

            sgi = self.get_cg_node_id(u)
            if not cg.has_node(sgi):
                label = self.get_sg_node_label(table, u)
                if isinstance(u, CellNode):
                    sgu = CGColumnNode(
                        id=sgi, label=label, column=u.column, nodes=set()
                    )
                elif isinstance(u, EntityValueNode):
                    sgu = CGEntityValueNode(
                        id=sgi,
                        label=label,
                        qnode_id=u.qnode_id,
                        context_span=u.context_span,
                    )
                else:
                    sgu = CGLiteralValueNode(
                        id=sgi, label=label, value=u.value, context_span=u.context_span
                    )
                cg.add_node(sgu)
            else:
                sgu = cg.get_node(sgi)

            if isinstance(sgu, CGColumnNode):
                sgu.nodes.add(u.id)

        # second step: add link
        # for uid, udata in dg.nodes(data=True):
        #     u: DGNode = udata["data"]
        for u in dg.iter_nodes():
            if isinstance(u, StatementNode):
                continue

            # add statement
            p2stmts: Dict[str, List[StatementNode]] = {}
            # for _, sid, us_eid, us_edata in dg.out_edges(uid, data=True, keys=True):
            for us_edge in dg.out_edges(u.id):
                if us_edge.predicate not in p2stmts:
                    p2stmts[us_edge.predicate] = []
                p2stmts[us_edge.predicate].append(dg.get_statement_node(us_edge.target))

            sgi = self.get_cg_node_id(u)
            for p, stmts in p2stmts.items():
                for stmt in stmts:
                    # we duplicate based on SG column node, not by DG node
                    children_values = {}
                    children_non_values = []

                    # for _, vid, sv_eid, sv_edata in dg.out_edges(
                    #     stmt.id, data=True, keys=True
                    # ):
                    #     v: DGNode = dg.nodes[vid]["data"]
                    for sv_edge in dg.out_edges(stmt.id):
                        v = dg.get_node(sv_edge.target)
                        if (v.id, sv_edge.predicate) not in stmt.forward_flow[u.id, p]:
                            # because of re-use statement, we need to only consider the children of this u node only
                            continue

                        if sv_edge.predicate == p:
                            tgi = self.get_cg_node_id(v)
                            if tgi not in children_values:
                                children_values[tgi] = []
                            children_values[tgi].append(v)
                        else:
                            children_non_values.append((sv_edge.predicate, v))

                    for tgi, vals in children_values.items():
                        stmt_id = CGStatementNode.get_id(sgi, p, tgi)
                        if not cg.has_node(stmt_id):
                            cg.add_node(CGStatementNode.new(stmt_id))

                        if not cg.has_edge_between_nodes(sgi, stmt_id, p):
                            cg.add_edge(
                                CGEdge(
                                    source=sgi,
                                    target=stmt_id,
                                    predicate=p,
                                    features={},
                                )
                            )
                        if not cg.has_edge_between_nodes(stmt_id, tgi, p):
                            cg.add_edge(
                                CGEdge(
                                    source=stmt_id,
                                    target=tgi,
                                    predicate=p,
                                    features={},
                                )
                            )

                        sg_stmt = cg.get_statement_node(stmt_id)
                        for v in vals:
                            sg_stmt.track_provenance(
                                CGEdgeFlowSource(u.id, sgi, p),
                                stmt.id,
                                CGEdgeFlowTarget(v.id, tgi, p),
                                stmt.get_provenance(
                                    EdgeFlowSource(u.id, p), EdgeFlowTarget(v.id, p)
                                ),
                            )

                        for qp, qv in children_non_values:
                            qtgi = self.get_cg_node_id(qv)
                            if not dg.has_edge_between_nodes(stmt_id, qtgi, qp):
                                cg.add_edge(
                                    CGEdge(
                                        source=stmt_id,
                                        target=qtgi,
                                        predicate=qp,
                                        features={},
                                    )
                                )

                            sg_stmt.track_provenance(
                                CGEdgeFlowSource(u.id, sgi, p),
                                stmt.id,
                                CGEdgeFlowTarget(qv.id, qtgi, qp),
                                stmt.get_provenance(
                                    EdgeFlowSource(u.id, p), EdgeFlowTarget(qv.id, qp)
                                ),
                            )
        M.log("grams", candidate_graph=cg)
        return cg

    @staticmethod
    def get_cg_node_id(u: DGNode):
        if isinstance(u, CellNode):
            return f"column-{u.column}"
        return u.id

    def get_sg_node_label(self, table: LinkedTable, u: DGNode):
        if isinstance(u, CellNode):
            return table.table.columns[u.column].name or ""
        if isinstance(u, EntityValueNode):
            qnode_label = self.wdentity_labels.get(u.qnode_id, u.qnode_id)
            return f"{qnode_label} ({u.qnode_id})"
        if isinstance(u, LiteralValueNode):
            return u.value.to_string_repr()
        raise M.UnreachableError(
            "This function doesn't supposed to be called with statement node"
        )

    def get_label(
        self,
        porq: str,
    ):
        """Get label for a property or qualifier"""
        if porq.startswith("Q"):
            if porq in self.wdclasses:
                item = self.wdclasses[porq]
            elif porq in self.wdentities:
                item = self.wdentities[porq]
            else:
                assert False
        else:
            item = self.wdprops[porq]

        return f"{item.label} ({item.id})"



from __future__ import annotations
from typing import Callable, Dict, Iterable, List, Mapping, Optional, Tuple, Literal
from grams.algorithm.candidate_graph.cg_graph import (
    CGColumnNode,
    CGEdgeTriple,
    CGGraph,
    CGStatementNode,
)
from grams.algorithm.data_graph.dg_graph import DGGraph
from grams.algorithm.inferences.features.structure_feature import StructureFeature
from grams.algorithm.inferences.psl_lib import (
    IDMap,
    PSLModel,
    RuleContainer,
)
from grams.algorithm.inferences.features.rel_feature import RelFeatures
from grams.algorithm.inferences.features.type_feature import (
    TypeFeatures,
)
from grams.inputs.linked_table import Link, LinkedTable
from kgdata.wikidata.models import (
    WDEntity,
    WDEntityLabel,
    WDClass,
    WDProperty,
    WDQuantityPropertyStats,
)
from pslpython.predicate import Predicate
from pslpython.model import Model
from dataclasses import dataclass

from pslpython.rule import Rule
from loguru import logger


class P:
    """Holding list of predicates in the model."""

    # target predicates
    CorrectRel = Predicate("CORRECT_REL", closed=False, size=3)
    CorrectType = Predicate("CORRECT_TYPE", closed=False, size=2)

    # graph structure
    Rel = Predicate("REL", closed=True, size=3)
    Type = Predicate("TYPE", closed=True, size=2)
    Statement = Predicate("STATEMENT", closed=True, size=1)

    # ontology
    SubProp = Predicate("SUB_PROP", closed=True, size=2)
    SubType = Predicate("SUB_TYPE", closed=True, size=2)
    HasSubType = Predicate("HAS_SUB_TYPE", closed=True, size=2)
    TypeDistance = Predicate("TYPE_DISTANCE", closed=True, size=2)
    NotRange = Predicate("NOT_RANGE", closed=True, size=2)
    DataProperty = Predicate("DATA_PROPERTY", closed=True, size=1)

    # features
    RelFreqOverRow = Predicate("REL_FREQ_OVER_ROW", closed=True, size=3)
    RelFreqOverEntRow = Predicate("REL_FREQ_OVER_ENT_ROW", closed=True, size=3)
    RelFreqOverPosRel = Predicate("REL_FREQ_OVER_POS_REL", closed=True, size=3)
    RelFreqUnmatchOverEntRow = Predicate(
        "REL_FREQ_UNMATCH_OVER_ENT_ROW", closed=True, size=3
    )
    RelFreqUnmatchOverPosRel = Predicate(
        "REL_FREQ_UNMATCH_OVER_POS_REL", closed=True, size=3
    )
    RelNotFuncDependency = Predicate("REL_NOT_FUNC_DEPENDENCY", closed=True, size=3)
    # RelIncorrectDataType = Predicate("REL_INCORRECT_DATA_TYPE", closed=True, size=3)
    RelHeaderSimilarity = Predicate("REL_HEADER_SIMILARITY", closed=True, size=3)

    TypeFreqOverRow = Predicate("TYPE_FREQ_OVER_ROW", closed=True, size=2)
    ExtendedTypeFreqOverRow = Predicate(
        "EXTENDED_TYPE_FREQ_OVER_ROW", closed=True, size=2
    )
    TypeHeaderSimilarity = Predicate("TYPE_HEADER_SIMILARITY", closed=True, size=2)


class PSLGramModel:
    def __init__(
        self,
        wdentities: Mapping[str, WDEntity],
        wdentity_labels: Mapping[str, WDEntityLabel],
        wdclasses: Mapping[str, WDClass],
        wdprops: Mapping[str, WDProperty],
        wd_numprop_stats: Mapping[str, WDQuantityPropertyStats],
        sim_fn: Optional[Callable[[str, str], float]] = None,
        disable_rules: Optional[Iterable[str]] = None,
    ):
        self.wdentities = wdentities
        self.wdentity_labels = wdentity_labels
        self.wdclasses = wdclasses
        self.wdprops = wdprops
        self.wd_numprop_stats = wd_numprop_stats
        self.sim_fn = sim_fn
        self.disable_rules = set(disable_rules or [])

        self.model = self.get_model()
        self.model.set_parameters(
            {
                P.Rel.name() + "_PRIOR_NEG": 1,
                "REL_PRIOR_NEG_PARENT_PROP_1": 0.1,
                "REL_PRIOR_NEG_PARENT_PROP_2": 0.1,
                "REL_PRIOR_NEG_PARENT_QUALIFIER": 0.1,
                P.RelFreqOverRow.name(): 2,
                P.RelFreqOverEntRow.name(): 2,
                P.RelFreqOverPosRel.name(): 2,
                P.RelFreqUnmatchOverEntRow.name(): 2,
                P.RelFreqUnmatchOverPosRel.name(): 2,
                P.RelHeaderSimilarity.name(): 2,
                # P.RelIncorrectDataType.name(): 2,
                P.RelNotFuncDependency.name(): 100,
                P.Type.name() + "_PRIOR_NEG": 1,
                "TYPE_PRIOR_NEG_PARENT": 0.1,
                "TYPE_PROP_RANGE": 2,
                P.ExtendedTypeFreqOverRow.name(): 2,
                P.TypeHeaderSimilarity.name(): 0.1,
                "CASCADING_ERROR_1": 2,
                "CASCADING_ERROR_2": 2,
                "CASCADING_ERROR_3": 2,
            }
        )

    def get_model(self):
        # ** DEFINE RULES **
        rules = RuleContainer()
        for feat in [
            P.RelFreqOverRow,
            P.RelFreqOverEntRow,
            P.RelFreqOverPosRel,
        ]:
            rules[feat.name()] = Rule(
                f"{P.Rel.name()}(N1, N2, P) & {feat.name()}(N1, N2, P) -> {P.CorrectRel.name()}(N1, N2, P)",
                weighted=True,
                squared=True,
                weight=0.0,
            )

        for feat in [
            P.RelFreqUnmatchOverEntRow,
            P.RelFreqUnmatchOverPosRel,
        ]:
            rules[feat.name()] = Rule(
                f"{P.Rel.name()}(N1, N2, P) & {feat.name()}(N1, N2, P) -> ~{P.CorrectRel.name()}(N1, N2, P)",
                weighted=True,
                squared=True,
                weight=0.0,
            )

        rules[P.RelNotFuncDependency.name()] = Rule(
            f"""
            {P.Rel.name()}(N1, N2, P) & 
            {P.Rel.name()}(N2, N3, P2) & 
            ~{P.Statement.name()}(N2) & 
            {P.CorrectRel.name()}(N1, N2, P) &
            {P.RelNotFuncDependency.name()}(N2, N3, P2) -> ~{P.CorrectRel.name()}(N2, N3, P2)
            """,
            weighted=True,
            squared=True,
            weight=0.0,
        )

        # prefer details prop/type
        # TODO: check REL_PRIOR_NEG_PARENT_PROP_1 whether we need them to connect
        # to the same V node
        rules["REL_PRIOR_NEG_PARENT_PROP_1"] = Rule(
            f"""
            {P.Rel.name()}(U, S, P) & {P.Statement.name()}(S) & {P.Rel.name()}(S, V, P) &
            {P.Rel.name()}(U, S2, PP) & {P.Statement.name()}(S2) & {P.Rel.name()}(S2, V, PP) & 
            {P.SubProp.name()}(P, PP) -> ~{P.CorrectRel.name()}(U, S2, PP)
            """,
            weighted=True,
            squared=True,
            weight=0.0,
        )
        rules["REL_PRIOR_NEG_PARENT_PROP_2"] = Rule(
            f"""
            {P.Rel.name()}(U, S, P) & {P.Statement.name()}(S) & {P.Rel.name()}(S, V, P) &
            {P.Rel.name()}(U, S2, PP) & {P.Statement.name()}(S2) & {P.Rel.name()}(S2, V, PP) &
            {P.SubProp.name()}(P, PP) -> ~{P.CorrectRel.name()}(S2, V, PP)
            """,
            weighted=True,
            squared=True,
            weight=0.0,
        )
        # rules["REL_PRIOR_NEG_PARENT_QUALIFIER"] = Rule(
        #     f"""
        #     {P.Rel.name()}(U, S, P) & {P.Statement.name()}(S) & {P.Rel.name()}(S, V, Q) & (P != Q) &
        #     {P.Rel.name()}(U, S2, PP) & {P.Statement.name()}(S2) & {P.Rel.name()}(S2, V, PQ) & (PQ != PP) &
        #     {P.SubProp.name()}(Q, PQ) -> ~{P.CorrectRel.name()}(S2, V, PQ)
        #     """,
        #     weighted=True,
        #     squared=True,
        #     weight=0.0,
        # )
        rules["TYPE_PRIOR_NEG_PARENT"] = Rule(
            f"{P.Type.name()}(N, T) & {P.TypeDistance.name()}(N, T) -> ~{P.CorrectType.name()}(N, T)",
            weighted=True,
            weight=0.0,
            squared=True,
        )

        # default negative rel/types
        rules[P.Rel.name() + "_PRIOR_NEG"] = Rule(
            f"~{P.CorrectRel.name()}(N1, N2, P)",
            weight=0.0,
            weighted=True,
            squared=True,
        )
        rules["TYPE_PRIOR_NEG"] = Rule(
            f"~{P.CorrectType.name()}(N, T)", weight=0.0, weighted=True, squared=True
        )

        rules["CASCADING_ERROR_1"] = Rule(
            f"""
            {P.Rel.name()}(U, S, P) & {P.Statement.name()}(S) & {P.Rel.name()}(S, V1, P) & 
            {P.Rel.name()}(S, V2, Q) & (V1 != V2) & ~{P.CorrectRel.name()}(S, V1, P) -> ~{P.CorrectRel.name()}(S, V2, Q)
            """,
            weighted=True,
            weight=0.0,
            squared=True,
        )
        rules["CASCADING_ERROR_2"] = Rule(
            f"""
            {P.Rel.name()}(U, S, P) & {P.Statement.name()}(S) & {P.Rel.name()}(S, V, P) & 
            ~{P.CorrectRel.name()}(S, V, P) -> ~{P.CorrectRel.name()}(U, S, P)""",
            weighted=True,
            weight=0.0,
            squared=True,
        )
        rules["CASCADING_ERROR_3"] = Rule(
            f"""
            {P.Rel.name()}(U, S, P) & {P.Statement.name()}(S) & {P.Rel.name()}(S, V, P) & 
            ~{P.CorrectRel.name()}(U, S, P) -> ~{P.CorrectRel.name()}(S, V, P)
            """,
            weighted=True,
            weight=0.0,
            squared=True,
        )

        for feat in [P.ExtendedTypeFreqOverRow, P.TypeHeaderSimilarity]:
            rules[feat.name()] = Rule(
                f"""
                {P.Type.name()}(N, T) & {feat.name()}(N, T) -> {P.CorrectType.name()}(N, T)
                """,
                weighted=True,
                weight=0.0,
                squared=True,
            )

        if self.disable_rules.difference(rules.keys()):
            raise Exception(
                f"Attempt to disable rules: {list(self.disable_rules.difference(rules.keys()))} that are not defined"
            )

        for rule in self.disable_rules:
            rules.pop(rule)

        return PSLModel(
            rules=rules,
            predicates=[x for x in vars(P).values() if isinstance(x, Predicate)],
            ignore_predicates_not_in_rules=True,
        )

    def predict(
        self,
        table: LinkedTable,
        cg: CGGraph,
        dg: DGGraph,
        verbose: bool = False,
        debug: bool = False,
    ) -> Tuple[Dict[CGEdgeTriple, float], Dict[int, Dict[str, float]]]:
        """Predict prob. of edges and prob. of columns' type"""
        idmap, observations, targets = self.extract_data(table, cg, dg)

        if verbose:
            for pname in observations:
                if len(observations[pname]) == 0:
                    logger.debug(f"No observations for predicate {pname}")

        output = self.model.predict(
            observations, targets, {}, force_setall=True, cleanup_tempdir=not debug
        )
        rel_probs = {
            tuple(idmap.im(t) for t in terms): prob
            for terms, prob in output[P.CorrectRel.name()].items()
        }

        type_probs = {}
        for terms, prob in output[P.CorrectType.name()].items():
            u = cg.get_node(idmap.im(terms[0]))
            class_id = idmap.im(terms[1])
            assert isinstance(u, CGColumnNode)
            type_probs.setdefault(u.column, {})[class_id] = prob

        if debug:
            self.model.debug(idmap)

        return rel_probs, type_probs

    def extract_data(self, table: LinkedTable, cg: CGGraph, dg: DGGraph):
        """Extract data for our PSL model"""
        idmap = IDMap()

        rel_feats = RelFeatures(
            idmap,
            table,
            cg,
            dg,
            self.wdentities,
            self.wdentity_labels,
            self.wdclasses,
            self.wdprops,
            self.wd_numprop_stats,
            self.sim_fn,
        ).extract_features(
            [
                P.RelFreqOverRow.name(),
                P.RelFreqOverEntRow.name(),
                P.RelFreqOverPosRel.name(),
                P.RelFreqUnmatchOverEntRow.name(),
                P.RelFreqUnmatchOverPosRel.name(),
                P.RelNotFuncDependency.name(),
            ]
        )

        type_feats = TypeFeatures(
            idmap,
            table,
            cg,
            dg,
            self.wdentities,
            self.wdclasses,
            self.wdprops,
            self.wd_numprop_stats,
            self.sim_fn,
        ).extract_features(
            [
                P.TypeFreqOverRow.name(),
                P.ExtendedTypeFreqOverRow.name(),
                P.TypeDistance.name(),
                P.HasSubType.name(),
            ]
        )

        candidate_types = {}
        for c, t, p in type_feats[P.TypeFreqOverRow.name()]:
            uid = idmap.im(c)
            if uid not in candidate_types:
                candidate_types[uid] = []
            candidate_types[uid].append(idmap.im(t))

        struct_feats = StructureFeature(
            idmap=idmap,
            table=table,
            cg=cg,
            dg=dg,
            wdentities=self.wdentities,
            wdentity_labels=self.wdentity_labels,
            wdclasses=self.wdclasses,
            wdprops=self.wdprops,
            wdprop_domains=None,
            wdprop_ranges=None,
            wd_num_prop_stats=self.wd_numprop_stats,
            sim_fn=self.sim_fn,
            candidate_types=candidate_types,
        ).extract_features(
            [
                P.Rel.name(),
                P.Type.name(),
                P.Statement.name(),
                P.SubProp.name(),
            ]
        )

        observations: Dict[str, list] = {}
        targets: Dict[str, list] = {}

        for p in self.model.model.get_predicates().values():
            p: Predicate
            if p.name() in rel_feats:
                observations[p.name()] = rel_feats[p.name()]
            if p.name() in type_feats:
                observations[p.name()] = type_feats[p.name()]
            if p.name() in struct_feats:
                observations[p.name()] = struct_feats[p.name()]

        # observations[P.RelHeaderSimilarity.name()] = []
        observations[P.TypeHeaderSimilarity.name()] = []

        targets[P.CorrectRel.name()] = observations[P.Rel.name()].copy()
        targets[P.CorrectType.name()] = observations[P.Type.name()].copy()

        # fitlering predicates that are not in the model (e.g. because they are not used in any rule)
        filtered_observations = {
            k: v for k, v in observations.items() if k in self.model.name2predicate
        }

        if len(filtered_observations) != len(observations):
            logger.warning(
                "The code is not efficient as it extracts more features than needed: {}",
                set(observations.keys()).difference(filtered_observations),
            )

        return idmap, filtered_observations, targets



import glob
from collections import defaultdict
from pathlib import Path
from typing import Dict, Optional, Tuple

import sm.misc as M
from grams.config import ROOT_DIR


def clean_table_linker(
    infile, outfile, ext_gt: Optional[Dict[Tuple[int, int], str]] = None
):
    rows = M.deserialize_csv(infile)
    cname2idx = {v: k for k, v in enumerate(rows[0])}

    cells = defaultdict(dict)
    for row in rows[1:]:
        ci = int(row[cname2idx["column"]])
        ri = int(row[cname2idx["row"]])
        if "GT_kg_id" in cname2idx:
            gt_ent = row[cname2idx["GT_kg_id"]]
        elif ext_gt is not None and (ri, ci) in ext_gt:
            gt_ent = ext_gt[(ri, ci)]
        else:
            gt_ent = None

        pred_ent = row[cname2idx["kg_id"]]
        score = float(row[cname2idx["siamese_prediction"]])

        if (ri, ci) in cells:
            assert cells[ri, ci]["gt"] == gt_ent
        cells[ri, ci]["gt"] = gt_ent

        if "pred_ent" not in cells[ri, ci]:
            cells[ri, ci]["pred_ent"] = []
        cells[ri, ci]["pred_ent"].append(f"{pred_ent}:{score:.9f}")

    links = []
    for (ri, ci), o in cells.items():
        links.append([ri, ci, o["gt"]] + o["pred_ent"])
    M.serialize_csv(links, outfile, delimiter="\t")


if __name__ == "__main__":
    for infile in glob.glob(str(ROOT_DIR / "examples/t2dv2/tables/*.table_linker.csv")):
        infile = Path(infile)
        outfile = infile.parent / f"{infile.name.split('.')[0]}.candidates.tsv"
        clean_table_linker(infile, outfile)



from __future__ import annotations

import copy
import re
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum, Flag, auto
from functools import cmp_to_key
from itertools import chain
from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple, Union, cast
from graph.retworkx import (
    RetworkXStrDiGraph,
    BaseEdge,
    BaseNode,
)

import networkx as nx
import sm.misc as M
from grams.algorithm.kg_index import KGObjectIndex
from grams.inputs.linked_table import LinkedTable
from kgdata.wikidata.models import WDValue, WDEntity, WDClass, WDProperty
from loguru import logger
from sm.misc.graph import viz_graph
from tqdm import tqdm
from typing_extensions import TypedDict


@dataclass(frozen=True)
class Span:
    start: int
    end: int

    @property
    def length(self):
        return self.end - self.start


@dataclass
class CellNode(BaseNode[str]):
    id: str
    value: str
    column: int
    row: int
    entity_ids: List[str]
    entity_spans: Dict[str, List[Span]]


@dataclass
class ContextSpan:
    # the text of the context
    text: str
    span: Span

    def get_text_span(self):
        return self.text[self.span.start : self.span.end]


@dataclass
class LiteralValueNode(BaseNode[str]):
    id: str
    value: WDValue
    # not none if it is appear in the context
    context_span: Optional[ContextSpan]

    @property
    def is_context(self):
        return self.context_span is not None


@dataclass
class EntityValueNode(BaseNode[str]):
    id: str
    qnode_id: str
    # not none if it is appear in the context
    context_span: Optional[ContextSpan]

    @property
    def is_context(self):
        return self.context_span is not None


# edge id is actually key id
EdgeFlowSource = NamedTuple("EdgeFlowSource", [("source_id", str), ("edge_id", str)])
EdgeFlowTarget = NamedTuple("EdgeFlowTarget", [("target_id", str), ("edge_id", str)])


@dataclass
class StatementNode(BaseNode[str]):
    id: str
    # id of the qnode that contains the statement
    qnode_id: str
    # predicate of the statement
    predicate: str
    # whether this statement actually exist in KG
    is_in_kg: bool

    # recording which link in the source is connected to the target.
    forward_flow: Dict[EdgeFlowSource, Set[EdgeFlowTarget]] = field(
        default_factory=dict
    )
    reversed_flow: Dict[EdgeFlowTarget, Set[EdgeFlowSource]] = field(
        default_factory=dict
    )
    flow: Dict[Tuple[EdgeFlowSource, EdgeFlowTarget], List[FlowProvenance]] = field(
        default_factory=dict
    )

    @property
    def value(self):
        return self.id

    def track_provenance(
        self,
        source_flow: EdgeFlowSource,
        target_flow: EdgeFlowTarget,
        provenances: List[FlowProvenance],
    ):
        ptr = self.forward_flow
        if source_flow not in ptr:
            ptr[source_flow] = set()
        ptr[source_flow].add(target_flow)

        ptr = self.reversed_flow
        if target_flow not in ptr:
            ptr[target_flow] = set()
        ptr[target_flow].add(source_flow)

        # TODO: merge the provenance if we can
        if (source_flow, target_flow) not in self.flow:
            self.flow[source_flow, target_flow] = provenances
        else:
            self.flow[source_flow, target_flow] = FlowProvenance.merge_lst(
                self.flow[source_flow, target_flow], provenances
            )

    def untrack_source_flow(self, source_flow: EdgeFlowSource):
        for target_flow in self.forward_flow.pop(source_flow):
            self.flow.pop((source_flow, target_flow))
            self.reversed_flow[target_flow].remove(source_flow)
            if len(self.reversed_flow[target_flow]) == 0:
                self.reversed_flow.pop(target_flow)

    def untrack_target_flow(self, target_flow: EdgeFlowTarget):
        for source_flow in self.reversed_flow.pop(target_flow):
            self.flow.pop((source_flow, target_flow))
            self.forward_flow[source_flow].remove(target_flow)
            if len(self.forward_flow[source_flow]) == 0:
                self.forward_flow.pop(source_flow)

    def untrack_flow(self, source_flow: EdgeFlowSource, target_flow: EdgeFlowTarget):
        self.flow.pop((source_flow, target_flow))
        self.forward_flow[source_flow].remove(target_flow)
        self.reversed_flow[target_flow].remove(source_flow)

        if len(self.forward_flow[source_flow]) == 0:
            self.forward_flow.pop(source_flow)
        if len(self.reversed_flow[target_flow]) == 0:
            self.reversed_flow.pop(target_flow)

    def has_source_flow(self, source_flow):
        return source_flow in self.forward_flow

    def has_target_flow(self, target_flow):
        return target_flow in self.reversed_flow

    def has_flow(self, source_flow, target_flow):
        return (source_flow, target_flow) in self.flow

    def iter_source_flow(self, target_flow: EdgeFlowTarget):
        for source_flow in self.reversed_flow[target_flow]:
            yield source_flow, self.flow[source_flow, target_flow]

    def iter_target_flow(self, source_flow: EdgeFlowSource):
        for target_flow in self.forward_flow[source_flow]:
            yield target_flow, self.flow[source_flow, target_flow]

    def get_provenance(self, source_flow: EdgeFlowSource, target_flow: EdgeFlowTarget):
        return self.flow[source_flow, target_flow]

    def get_provenance_by_edge(self, inedge: "DGEdge", outedge: "DGEdge"):
        return self.flow[
            EdgeFlowSource(inedge.source, inedge.predicate),
            EdgeFlowTarget(outedge.target, outedge.predicate),
        ]

    def is_same_flow(self, inedge: "DGEdge", outedge: "DGEdge") -> bool:
        return (
            EdgeFlowSource(inedge.source, inedge.predicate),
            EdgeFlowTarget(outedge.target, outedge.predicate),
        ) in self.flow


DGNode = Union[CellNode, LiteralValueNode, EntityValueNode, StatementNode]


class LinkGenMethod(Enum):
    """Methods to generate a link"""

    # this statement is generated by exact matching from the link
    FromWikidataLink = "from_wikidata_link"
    FromLiteralMatchingFunc = "from_literal_matching_function"


@dataclass
class FlowProvenance:
    """Contain information regarding how this relationship/flow has been generated (typically coming from the matching algorithm)"""

    # method that
    gen_method: LinkGenMethod
    gen_method_arg: Any
    prob: float

    def merge(self, another: FlowProvenance) -> Union[FlowProvenance, None]:
        """Try to merge the provenance, if we cannot merge them, just return None"""
        if self.gen_method != another.gen_method:
            return None
        if self.gen_method == LinkGenMethod.FromWikidataLink:
            return self
        if self.gen_method == LinkGenMethod.FromLiteralMatchingFunc:
            if self.gen_method_arg == another.gen_method_arg:
                if self.prob > another.prob:
                    return self
                return another
            return None
        raise M.UnreachableError()

    @staticmethod
    def merge_lst(
        lst1: List[FlowProvenance], lst2: List[FlowProvenance]
    ) -> List[FlowProvenance]:
        """Assume that items within each list (lst1 & lst2) are not mergeable"""
        if len(lst1) == 0:
            return lst2
        elif len(lst2) == 0:
            return lst1

        lst = copy.copy(lst1)
        for item in lst2:
            for i in range(len(lst1)):
                resp = lst[i].merge(item)
                if resp is None:
                    # cannot merge them
                    lst.append(item)
                    break
                else:
                    # merge lst[i] & item
                    lst[i] = resp
                    break
        return lst


@dataclass
class DGStatementID:
    qnode_id: str
    predicate: str
    statement_index: int

    def get_id(self):
        return f"stmt:{self.qnode_id}-{self.predicate}-{self.statement_index}"

    @staticmethod
    def parse_id(id: str) -> "DGStatementID":
        m = re.match(r"stmt:([^-]+)-([^-]+)-([^-]+)", id)
        assert m is not None
        qnode_id, predicate, stmt_index = [m.group(i) for i in range(1, 4)]
        stmt_index = int(stmt_index)
        return DGStatementID(qnode_id, predicate, stmt_index)


@dataclass
class DGPathNodeStatement:
    qnode_id: str
    predicate: str
    statement_index: int
    # how this DGPath statement has been matched
    provenance: FlowProvenance

    def get_id(self):
        return DGStatementID(
            self.qnode_id, self.predicate, self.statement_index
        ).get_id()

    # ############# METHODs to construct DGPathNodeStatement from provenance ##############
    @staticmethod
    def from_FromWikidataLink(qnode_id, predicate, stmt_index):
        return DGPathNodeStatement(
            qnode_id,
            predicate,
            stmt_index,
            FlowProvenance(LinkGenMethod.FromWikidataLink, None, 1.0),
        )

    @staticmethod
    def from_FromLiteralMatchingFunc(qnode_id, predicate, stmt_index, fn_args, prob):
        return DGPathNodeStatement(
            qnode_id,
            predicate,
            stmt_index,
            FlowProvenance(LinkGenMethod.FromLiteralMatchingFunc, fn_args, prob),
        )


@dataclass
class DGPathNodeEntity:
    qnode_id: str

    def get_id(self):
        return f"ent:{self.qnode_id}"


@dataclass
class DGPathNodeLiteralValue:
    value: WDValue

    def get_id(self):
        return f"val:{self.value.to_string_repr()}"


@dataclass
class DGPathExistingNode:
    id: str

    def get_id(self):
        return self.id


DGPathNode = Union[
    DGPathNodeStatement, DGPathNodeEntity, DGPathNodeLiteralValue, DGPathExistingNode
]


@dataclass
class DGPathEdge:
    value: str
    is_qualifier: bool

    @staticmethod
    def p(value: str):
        return DGPathEdge(value, is_qualifier=False)

    @staticmethod
    def q(value: str):
        return DGPathEdge(value, is_qualifier=True)


@dataclass
class DGPath:
    # a sequence of path, is always
    sequence: List[Union[DGPathEdge, DGPathNode]] = field(default_factory=list)


@dataclass
class DGEdge(BaseEdge[str, str]):
    source: str
    target: str
    predicate: str
    is_qualifier: bool
    # deprecated, will be replaced by the information stored directly in the statement edge flow
    # paths: List[DGPath]
    is_inferred: Optional[bool] = None
    id: int = -1  # set automatically by the graph

    @property
    def key(self):
        return self.predicate

    @staticmethod
    def can_link(source: DGNode, target: DGNode):
        """Test if there can be two links between the two nodes. Basically, it just enforce the constraint that no link
        cross different rows
        """
        if not isinstance(source, CellNode) or not isinstance(target, CellNode):
            return True
        return source.row == target.row


class DGGraph(RetworkXStrDiGraph[str, DGNode, DGEdge]):
    def get_statement_node(self, nid: str) -> StatementNode:
        n = self.get_node(nid)
        assert isinstance(n, StatementNode)
        return n

    def get_cell_node(self, nid: str) -> CellNode:
        n = self.get_node(nid)
        assert isinstance(n, CellNode)
        return n
